import sys
from typing import TYPE_CHECKING, Optional

from qualia_core.experimenttracking.NeptuneBase import NeptuneBase
from qualia_core.typing import RecursiveConfigDict

if sys.version_info >= (3, 12):
    from typing import override
else:
    from typing_extensions import override

if TYPE_CHECKING:
    from neptune.new.integrations.tensorflow_keras import NeptuneCallback

class Neptune(NeptuneBase):
    def __init__(self, project_name: str, config_file: str='conf/neptune.toml') -> None:
        super().__init__(project_name=project_name, config_file=config_file)

    @override
    def start(self, name: Optional[str] = None) -> None:
        import neptune.new as neptune
        from neptune.new.integrations.tensorflow_keras import NeptuneCallback

        project_name = f'{self.project_name}_{name}' if name is not None else self.project_name

        self.run = neptune.init(
            api_key=self.api_key,
            project_name=f'{self.project_namespace}/{project_name}',
            upload_source_files=self.source_files)
        self.neptune_cbk = NeptuneCallback(run=self.run, base_namespace='metrics')

    @override
    def stop(self) -> None:
        self.run.stop()

    @property
    def callback(self) -> 'NeptuneCallback':
        return self.neptune_cbk

    @override
    def _hyperparameters(self, params: RecursiveConfigDict) -> None:
        self.neptune_cbk.log_hyperparams(params)



from pathlib import Path
from typing import Callable

import pytest


class TestQualiaCodeGenSparkFunEdge:
    '''Must be executed from command line with PYTHONHASHSEED environment variable defined to the "seed" value of the "bench"
    section of the configuration file.
    '''

    @pytest.mark.dependency()
    def test_uci_har_resnet_qualia_codegen_sparkfunedge_float32(self, fixture_uci_har_resnet_train_float32: Callable[[], dict]) -> None:
        from qualia_core import main
        import qualia_core.utils.config
        configf, configname = qualia_core.utils.config.parse_config(Path('conf')/'tests'/'UCI-HAR_ResNetv1_QualiaCodeGen_SparkFunEdge_float32.toml')
        config = qualia_core.utils.config.validate_config_dict(configf)
        loggers = fixture_uci_har_resnet_train_float32

        assert 'learningmodel' in loggers

        assert len(loggers['learningmodel']) == 1
        assert loggers['learningmodel'][0].name == 'uci-har_resnetv1_8'
        assert loggers['learningmodel'][0].i == 1
        assert loggers['learningmodel'][0].params == 1150
        assert loggers['learningmodel'][0].mem_params == 4600
        assert loggers['learningmodel'][0].accuracy >= 0.85

        loggers = {**main.qualia('prepare_deploy', config, configname), **loggers}

        assert 'prepare_deploy' in loggers

        assert len(loggers['prepare_deploy']) == 1
        assert loggers['prepare_deploy'][0].name == 'uci-har_resnetv1_8'
        assert loggers['prepare_deploy'][0].quantize == 'float32'
        assert loggers['prepare_deploy'][0].optimize == ''
        assert loggers['prepare_deploy'][0].compress == 1

    @pytest.mark.dependency(depends=['TestQualiaCodeGenSparkFunEdge::test_uci_har_resnet_qualia_codegen_sparkfunedge_float32'])
    @pytest.mark.deploy()
    def test_uci_har_resnet_qualia_codegen_sparkfunedge_float32_deploy_and_evaluate(self):
        from qualia_core import main
        import qualia_core.utils.config
        configf, configname = qualia_core.utils.config.parse_config(Path('conf')/'tests'/'UCI-HAR_ResNetv1_QualiaCodeGen_SparkFunEdge_float32.toml')
        config = qualia_core.utils.config.validate_config_dict(configf)
        loggers = main.qualia('deploy_and_evaluate', config, configname)

        assert 'evaluate' in loggers

        assert len(loggers['evaluate']) == 1
        assert loggers['evaluate'][0].name == 'uci-har_resnetv1_8'
        assert loggers['evaluate'][0].i == 1
        assert loggers['evaluate'][0].quantization == 'float32'
        assert loggers['evaluate'][0].params == 1150
        assert loggers['evaluate'][0].mem_params == 4600
        assert loggers['evaluate'][0].accuracy >= 0.85
        assert loggers['evaluate'][0].avg_time <= 0.03
        assert loggers['evaluate'][0].rom_size <= 45000

    @pytest.mark.dependency()
    def test_uci_har_resnet_qualia_codegen_sparkfunedge_int16(self, fixture_uci_har_resnet_train_int16: Callable[[], dict]) -> None:
        from qualia_core import main
        import qualia_core.utils.config
        configf, configname = qualia_core.utils.config.parse_config(Path('conf')/'tests'/'UCI-HAR_ResNetv1_QualiaCodeGen_SparkFunEdge_int16.toml')
        config = qualia_core.utils.config.validate_config_dict(configf)
        loggers = fixture_uci_har_resnet_train_int16

        assert 'learningmodel' in loggers

        assert len(loggers['learningmodel']) == 2 # float32 and int8 models
        assert loggers['learningmodel'][0].name == 'uci-har_resnetv1_8'
        assert loggers['learningmodel'][1].name == 'uci-har_resnetv1_8_q16_force_q9_e0_LSQFalse'
        assert loggers['learningmodel'][0].i == 1
        assert loggers['learningmodel'][1].i == 1
        assert loggers['learningmodel'][0].params == 1150
        assert loggers['learningmodel'][1].params == 1150
        assert loggers['learningmodel'][0].mem_params == 4600 # Model is float32 after training
        assert loggers['learningmodel'][1].mem_params == 2300 # Model is (virtually) int8 after QAT
        assert loggers['learningmodel'][0].accuracy >= 0.85
        assert loggers['learningmodel'][1].accuracy >= 0.85

        loggers = {**main.qualia('prepare_deploy', config, configname), **loggers}

        assert 'prepare_deploy' in loggers

        assert len(loggers['prepare_deploy']) == 1
        assert loggers['prepare_deploy'][0].name == 'uci-har_resnetv1_8_q16_force_q9_e0_LSQFalse'
        assert loggers['prepare_deploy'][0].quantize == 'int16'
        assert loggers['prepare_deploy'][0].optimize == ''
        assert loggers['prepare_deploy'][0].compress == 1

    @pytest.mark.dependency(depends=['TestQualiaCodeGenSparkFunEdge::test_uci_har_resnet_qualia_codegen_sparkfunedge_int16'])
    @pytest.mark.deploy()
    def test_uci_har_resnet_qualia_codegen_sparkfunedge_int16_deploy_and_evaluate(self):
        from qualia_core import main
        import qualia_core.utils.config
        configf, configname = qualia_core.utils.config.parse_config(Path('conf')/'tests'/'UCI-HAR_ResNetv1_QualiaCodeGen_SparkFunEdge_int16.toml')
        config = qualia_core.utils.config.validate_config_dict(configf)
        loggers = main.qualia('deploy_and_evaluate', config, configname)

        assert 'evaluate' in loggers

        assert len(loggers['evaluate']) == 1
        assert loggers['evaluate'][0].name == 'uci-har_resnetv1_8_q16_force_q9_e0_LSQFalse'
        assert loggers['evaluate'][0].i == 1
        assert loggers['evaluate'][0].quantization == 'int16'
        assert loggers['evaluate'][0].params == 1150
        assert loggers['evaluate'][0].mem_params == 2300 # But quantized during deployment and int16 for evaluate
        assert loggers['evaluate'][0].accuracy >= 0.85
        assert loggers['evaluate'][0].avg_time <= 0.02
        assert loggers['evaluate'][0].rom_size <= 43000

    @pytest.mark.dependency()
    def test_uci_har_resnet_qualia_codegen_sparkfunedge_int8(self, fixture_uci_har_resnet_train_int8: Callable[[], dict]) -> None:
        from qualia_core import main
        import qualia_core.utils.config
        configf, configname = qualia_core.utils.config.parse_config(Path('conf')/'tests'/'UCI-HAR_ResNetv1_QualiaCodeGen_SparkFunEdge_int8.toml')
        config = qualia_core.utils.config.validate_config_dict(configf)
        loggers = fixture_uci_har_resnet_train_int8

        assert 'learningmodel' in loggers

        assert len(loggers['learningmodel']) == 2 # float32 and int8 models
        assert loggers['learningmodel'][0].name == 'uci-har_resnetv1_8'
        assert loggers['learningmodel'][1].name == 'uci-har_resnetv1_8_q8_force_qoff_e1_LSQFalse'
        assert loggers['learningmodel'][0].i == 1
        assert loggers['learningmodel'][1].i == 1
        assert loggers['learningmodel'][0].params == 1150
        assert loggers['learningmodel'][1].params == 1150
        assert loggers['learningmodel'][0].mem_params == 4600 # Model is float32 after training
        assert loggers['learningmodel'][1].mem_params == 1150 # Model is (virtually) int8 after QAT
        assert loggers['learningmodel'][0].accuracy >= 0.85
        assert loggers['learningmodel'][1].accuracy >= 0.85

        loggers = {**main.qualia('prepare_deploy', config, configname), **loggers}

        assert 'prepare_deploy' in loggers

        assert len(loggers['prepare_deploy']) == 1
        assert loggers['prepare_deploy'][0].name == 'uci-har_resnetv1_8_q8_force_qoff_e1_LSQFalse'
        assert loggers['prepare_deploy'][0].quantize == 'int8'
        assert loggers['prepare_deploy'][0].optimize == ''
        assert loggers['prepare_deploy'][0].compress == 1

    @pytest.mark.dependency(depends=['TestQualiaCodeGenSparkFunEdge::test_uci_har_resnet_qualia_codegen_sparkfunedge_int8'])
    @pytest.mark.deploy()
    def test_uci_har_resnet_qualia_codegen_sparkfunedge_int8_deploy_and_evaluate(self):
        from qualia_core import main
        import qualia_core.utils.config
        configf, configname = qualia_core.utils.config.parse_config(Path('conf')/'tests'/'UCI-HAR_ResNetv1_QualiaCodeGen_SparkFunEdge_int8.toml')
        config = qualia_core.utils.config.validate_config_dict(configf)
        loggers = main.qualia('deploy_and_evaluate', config, configname)

        assert 'evaluate' in loggers

        assert len(loggers['evaluate']) == 1
        assert loggers['evaluate'][0].name == 'uci-har_resnetv1_8_q8_force_qoff_e1_LSQFalse'
        assert loggers['evaluate'][0].i == 1
        assert loggers['evaluate'][0].quantization == 'int8'
        assert loggers['evaluate'][0].params == 1150
        assert loggers['evaluate'][0].mem_params == 1150 # Quantized during deployment and int8 for evaluate
        assert loggers['evaluate'][0].accuracy >= 0.85
        assert loggers['evaluate'][0].avg_time <= 0.02
        assert loggers['evaluate'][0].rom_size <= 39000



# TO DO : Delete

from typing import Optional

import torch


class IntNoGradient(torch.autograd.Function):
    @staticmethod
    def forward(ctx, x):
        return x.floor()

    @staticmethod
    def backward(ctx, g):
        return g

def maxWeight(weight: torch.Tensor, maxif: float, training: bool) -> torch.Tensor:
    maxi = torch.tensor(maxif, device=weight.device)

#    if(maxi==0 or abs(maxi) == float('inf') or training):
#        maxi = torch.max(torch.abs(weight.view(-1)))

    if maxi == 0:
        maxi = torch.tensor(-1, device=weight.device)
    elif abs(maxi) != float('inf'):
        maxi = -(torch.log2(maxi).floor().int() + 1)
    else:
        maxi = torch.tensor(0, device=weight.device)
    return maxi


def quantifier(w: torch.Tensor, n_bit: int, training: bool, maxi: float = 0, force_q: Optional[int] = None) -> torch.Tensor:
    tn_bit = torch.tensor(n_bit, device=w.device)

    if force_q is not None:
        v = torch.pow(2.0, torch.tensor(force_q, device=w.device))
    else:
        qint = maxWeight(w, maxi, training)
        v = torch.pow(2.0, tn_bit - 1 + qint)

    w = w*v
    w = IntNoGradient.apply(w)
    w = torch.clamp(w, -torch.pow(2, tn_bit - 1), torch.pow(2, tn_bit - 1) - 1)
    w = w / v

    return w



from __future__ import annotations

import sys

from qualia_core.typing import TYPE_CHECKING

from .DataAugmentationPyTorch import DataAugmentationPyTorch

if TYPE_CHECKING:
    import torch  # noqa: TCH002

if sys.version_info >= (3, 12):
    from typing import override
else:
    from typing_extensions import override

class Rotation2D(DataAugmentationPyTorch):

    def __init__(self,
                 angle: list[int] | None = None,
                 evaluate: bool = False,  # noqa: FBT001, FBT002
                 before: bool = False,  # noqa: FBT001, FBT002
                 after: bool = True) -> None:  # noqa: FBT001, FBT002
        import torchvision.transforms  # type: ignore[import-untyped]

        super().__init__(evaluate=evaluate, before=before, after=after)

        angle = angle if angle is not None else [-45, 45]
        self.randomrotation = torchvision.transforms.RandomRotation(angle)

    def apply(self, x: torch.Tensor, y: torch.Tensor) -> tuple[torch.Tensor, torch.Tensor]:
        rotated_x = self.randomrotation(x)

        return rotated_x, y

    @override
    def __call__(self, data: tuple[torch.Tensor, torch.Tensor], device: torch.device) -> tuple[torch.Tensor, torch.Tensor]:
        return self.apply(*data)



from __future__ import annotations

import logging
import os
import stat
import sys
from pathlib import Path
from typing import Any

from qualia_core.deployment.toolchain import STM32CubeIDE
from qualia_core.evaluation.target.STM32CubeAI import STM32CubeAI as STM32CubeAIEvaluator
from qualia_core.typing import TYPE_CHECKING

if TYPE_CHECKING:
    if sys.version_info >= (3, 11):
        from typing import Self
    else:
        from typing_extensions import Self

    from qualia_core.postprocessing.Converter import Converter  # noqa: TCH001
    from qualia_core.postprocessing.Keras2TFLite import Keras2TFLite  # noqa: TCH001

if sys.version_info >= (3, 12):
    from typing import override
else:
    from typing_extensions import override

logger = logging.getLogger(__name__)

class STM32CubeAI(STM32CubeIDE):
    evaluator = STM32CubeAIEvaluator # Suggested evaluator

    def __init__(self,
                 projectname: str,
                 projectdir: Path,
                 outdir: Path | None = None) -> None:
        super().__init__(projectname=projectname, projectdir=projectdir, outdir=outdir)

        # Project not compatible with 6.0.0 yet, and 6.0.0 is buggy anyway. Use 5.2.0
        self.__stm32cubeai_bin = Path.home()/'STM32Cube'/'Repository'/'Packs'/'STMicroelectronics'/'X-CUBE-AI'/'5.2.0'/'Utilities'/'linux'/'stm32ai'
        #self.__stm32cubeai_bin = next((Path.home()/'STM32Cube'/'Repository'/'Packs'/'STMicroelectronics'/'X-CUBE-AI').glob('*'))/'Utilities'/'linux'/'stm32ai'

    def _create_modeloutdir(self, modelpath: Path) -> None:
        modelpath.mkdir(parents=True, exist_ok=True)

    def __write_model(self, model: Keras2TFLite, modelpath: Path) -> None:
        with modelpath.open('wb') as f:
            _ = f.write(model.data)

    def __generate(self, modelpath: Path, compression: int) -> bool:
        if compression != 1:
            logger.error('FIXME: Evaluation not logging compression level != 1')
            raise ValueError
        if compression not in [1, 4, 8]:
            logger.error('Compression factor %s is not supported, must be either 1, 4 or 8', compression)
            raise ValueError
        # Sometimes the executable may not have have exec permission, set it
        if not os.access(self.__stm32cubeai_bin, os.X_OK):
            self.__stm32cubeai_bin.chmod(self.__stm32cubeai_bin.stat().st_mode | stat.S_IXUSR)

        return self._run(self.__stm32cubeai_bin,
                            'generate',
                            '--model', str(modelpath),
                            '--output', str(self._projectdir/'X-CUBE-AI'/'App'),
                            '--compression', str(compression),
                            '--workspace', str(self._workspacedir),
                        )

    @override
    def prepare(self,
                tag: str,
                model: Converter[Any],
                optimize: str,
                compression: int) -> Self | None:
        # Keep here for isinstance() to avoid circual import
        from qualia_core.postprocessing.Keras2TFLite import Keras2TFLite

        if optimize != 'cmsis-nn':
            logger.error('cmsis-nn optimize mandatory for %s', type(self).__name__)
            raise ValueError

        if not isinstance(model, Keras2TFLite):
            logger.error('%s excepts the model to come from a Keras2TFLite Converter', type(self).__name__)
            raise TypeError


        modeloutdir = Path('out')/'deploy'/'stm32cubeai'
        modelpath = modeloutdir/f'{tag}.tflite'
        self._create_modeloutdir(modeloutdir)
        self._create_outdir()
        self._clean_workspace()
        self.__write_model(model=model, modelpath=modelpath)

        if not self.__generate(modelpath=modelpath, compression=compression):
            return None
        if not self._build():
            return None
        self._copy_buildproduct(tag=tag)
        return self



from .Accelerometer import Accelerometer
from .Barometer import Barometer
from .GPS import GPS
from .Gyroscope import Gyroscope
from .Magnetometer import Magnetometer
from .Sensor import Sensor

__all__ = ['Accelerometer', 'Barometer', 'GPS', 'Gyroscope', 'Magnetometer', 'Sensor']



import torch

import math



def minmax_range_setting(w: torch.Tensor, is_asymmetric: bool = False) -> tuple[torch.Tensor, torch.Tensor]:

    if is_asymmetric:
        max, min = torch.max(w), torch.min(w)
    else:
        max, min = (torch.max(torch.abs(w))), -(torch.max(torch.abs(w)))

    return max, min

def MSE_analysis_range_setting(w: torch.Tensor, n_bit: int, is_asymmetric: bool = False) -> tuple[torch.Tensor, torch.Tensor]:

    Alpha = torch.arange(1, 80, 0.01)
    std = torch.std(w)

    analysis = SymGaussianClippingAnalysis(w, Alpha, std, n_bit)

    winner_alpha = Alpha[torch.argmin(torch.tensor(analysis))] 

    max = (2*winner_alpha)
    min = -(2*winner_alpha) # to do for asymetric

    return torch.tensor(max), torch.tensor(min)

def MSE_simulation_range_setting(w: torch.Tensor, n_bit: int, is_asymmetric: bool = False) -> tuple[torch.Tensor, torch.Tensor]:

    Alpha = torch.arange(1, 40, 0.1)
    std = torch.std(w)

    simulation = SymGaussianClippingSimulation(w, Alpha, std, n_bit)

    winner_alpha = Alpha[torch.argmin(torch.tensor(simulation))] 

    max = (2*winner_alpha)
    min = -(2*winner_alpha) # to do for asymetric

    return torch.tensor(max), torch.tensor(min)

def uniform_quantizer(x, S):
    xQ = torch.round(x / S) * S
    return xQ

# TO DO
def SymGaussianClippingAnalysis(x_float, Alpha, sigma,bitWidth):
    Analysis = []
    for alpha in Alpha:
        clipping_mse = (sigma**2 + (alpha ** 2)) * (1 - math.erf(alpha / (sigma*torch.sqrt(torch.tensor(2.0))))) - torch.sqrt(torch.tensor(2.0/torch.pi)) * alpha * sigma*(torch.e ** ((-1)*(0.5* (alpha ** 2))/sigma**2))
        quant_mse = (alpha ** 2) / (3 * (2 ** (2 * bitWidth)))
        mse = clipping_mse + quant_mse
        Analysis.append(mse)
    return Analysis

def SymGaussianClippingSimulation(x_float, Alpha, sigma,bitWidth):
    simulations = []
    for alpha in Alpha:
        s = torch.clone(x_float)
        S = (2*alpha)/(2**bitWidth)

        # clipping
        s[s > alpha] = alpha
        s[s < -alpha] = -alpha

        # quantization
        s = uniform_quantizer(s, S)

        mse = ((s - x_float) ** 2).mean()
        simulations.append(mse)
    return simulations


class DataAugmentation:
    def __init__(self, evaluate: bool = False, before: bool = False, after: bool = True) -> None:  # noqa: FBT001, FBT002
        super().__init__()
        self.evaluate = evaluate # Whether to enable data augmentation for evaluation as wel
        self.before = before
        self.after = after



from __future__ import annotations

import sys

import torch
import torch.nn
from torch import nn

from qualia_core.learningmodel.pytorch.layers.QuantizedLayer import (
    QuantizedLayer,
    QuantizerActProtocol,
    QuantizerInputProtocol,
    QuantizerWProtocol,
)

from .layers.quantized_layers import QuantizedBatchNorm
from .Quantizer import QuantizationConfig, Quantizer, update_params

if sys.version_info >= (3, 12):
    from typing import override
else:
    from typing_extensions import override

class QuantizedConv1d(nn.Conv1d, QuantizerInputProtocol, QuantizerActProtocol, QuantizerWProtocol, QuantizedLayer):
    def __init__(self,  # noqa: PLR0913
                 in_channels: int,
                 out_channels: int,
                 quant_params: QuantizationConfig,
                 kernel_size: int = 3,
                 stride: int = 1,
                 padding: int = 0,
                 dilation: int = 1,
                 groups: int = 1,
                 bias: bool = True,  # noqa: FBT001, FBT002
                 activation: nn.Module | None = None) -> None:
        self.call_super_init = True # Support multiple inheritance from nn.Module
        super().__init__(in_channels,
                         out_channels,
                         kernel_size=kernel_size,
                         stride=stride,
                         padding=padding,
                         dilation=dilation,
                         groups=groups,
                         bias=bias)
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.kernel_width = kernel_size
        self.groups = groups
        self.activation = activation
        # Create the quantizer instance
        quant_params_input = update_params(tensor_type='input', quant_params=quant_params)
        quant_params_act = update_params(tensor_type='act', quant_params=quant_params)
        quant_params_w = update_params(tensor_type='w', quant_params=quant_params)
        self.quantizer_input = Quantizer(**quant_params_input)
        self.quantizer_act = Quantizer(**quant_params_act)
        self.quantizer_w = Quantizer(**quant_params_w)
        if 'bias' in quant_params :
            quant_params_bias = update_params(tensor_type='bias', quant_params=quant_params)
            self.quantizer_bias = Quantizer(**quant_params_bias)


    @override
    def forward(self, input: torch.Tensor) -> torch.Tensor:  # noqa: A002
        q_input = self.quantizer_input(input)

        if self.bias is None :
            # If no bias, quantize weights only
            q_w = self.quantizer_w(self.weight, bias_tensor=self.bias)
            y = torch.nn.functional.conv1d(q_input,
                                           q_w,
                                           stride=self.stride,
                                           padding=self.padding,
                                           dilation = self.dilation,
                                           groups=self.groups)
        else :
            # Quantize bias and weights
            if hasattr(self, 'quantizer_bias'):
                #...with the different quantization schemes
                q_w = self.quantizer_w(self.weight)
                q_b = self.quantizer_bias(self.bias)
            else :
                #...with the same quantization schemes
                q_w, q_b = self.quantizer_w(self.weight, bias_tensor=self.bias)
            y = torch.nn.functional.conv1d(q_input,
                                           q_w,
                                           bias=q_b,
                                           stride=self.stride,
                                           padding=self.padding,
                                           dilation=self.dilation,
                                           groups=self.groups)

        if self.activation:
            y = self.activation(y)

        return self.quantizer_act(y)


class QuantizedMaxPool1d(nn.MaxPool1d, QuantizerInputProtocol, QuantizerActProtocol, QuantizedLayer):
    def __init__(self,  # noqa: PLR0913
                 kernel_size: int,
                 quant_params: QuantizationConfig,
                 stride: int | None = None,
                 padding: int = 0,
                 activation: nn.Module | None = None) -> None:
        self.call_super_init = True # Support multiple inheritance from nn.Module
        super().__init__(kernel_size, stride=stride, padding=padding)
        self.activation = activation
        quant_params_input = update_params(tensor_type= 'input', quant_params=quant_params)
        quant_params_act = update_params(tensor_type='act', quant_params=quant_params)
        self.quantizer_input = Quantizer(**quant_params_input)
        self.quantizer_act = Quantizer(**quant_params_act)

    @override
    def forward(self, input: torch.Tensor) -> torch.Tensor:  # noqa: A002

        q_input = self.quantizer_input(input)

        y: torch.Tensor = super().forward(q_input)

        if self.activation:
            y = self.activation(y)

        return self.quantizer_act(y)


class QuantizedAdaptiveAvgPool1d(torch.nn.AdaptiveAvgPool1d, QuantizerInputProtocol, QuantizerActProtocol, QuantizedLayer):
    def __init__(self,
                 output_size: int,
                 quant_params: QuantizationConfig,
                 activation: nn.Module | None = None) -> None:
        self.call_super_init = True # Support multiple inheritance from nn.Module
        super().__init__(output_size)
        self.activation = activation
        quant_params_input = update_params(tensor_type='input', quant_params=quant_params)
        quant_params_act = update_params(tensor_type='act', quant_params=quant_params)
        self.quantizer_input = Quantizer(**quant_params_input)
        self.quantizer_act = Quantizer(**quant_params_act)

    @override
    def forward(self, input: torch.Tensor) -> torch.Tensor:  # noqa: A002
        q_input = self.quantizer_input(input)

        y = super().forward(q_input)

        if self.activation:
            y = self.activation(y)

        return self.quantizer_act(y)


class QuantizedBatchNorm1d(QuantizedBatchNorm):
    ...


class QuantizedAvgPool1d(torch.nn.AvgPool1d, QuantizerInputProtocol, QuantizerActProtocol, QuantizedLayer):
    def __init__(self,
                 kernel_size: int | tuple[int],
                 quant_params: QuantizationConfig,
                 stride: int | tuple[int] | None = None,
                 padding: int | tuple[int] = 0,
                 activation: nn.Module | None = None) -> None:
        self.call_super_init = True # Support multiple inheritance from nn.Module
        super().__init__(kernel_size, stride=stride, padding=padding)
        self.activation = activation
        quant_params_input = update_params(tensor_type='input', quant_params=quant_params)
        quant_params_act = update_params(tensor_type='act', quant_params=quant_params)
        self.quantizer_input = Quantizer(**quant_params_input)
        self.quantizer_act = Quantizer(**quant_params_act)

    @override
    def forward(self, input: torch.Tensor) -> torch.Tensor:  # noqa: A002
        q_input = self.quantizer_input(input)

        y = super().forward(q_input)

        if self.activation:
            y = self.activation(y)

        return self.quantizer_act(y)



from __future__ import annotations

import sys
from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING, Final

from qualia_core.datamodel import TimeSample
from qualia_core.datamodel.har import Activities, Activity, HARDataModel, Subject
from qualia_core.datamodel.sensor import Accelerometer, Barometer, Gyroscope
from qualia_core.utils.file import CSVReader, DirectoryReader

from .HARDataset import HARDataset

if TYPE_CHECKING:
    from qualia_core.datamodel.sensor.Sensor import Sensor

if sys.version_info >= (3, 12):
    from typing import override
else:
    from typing_extensions import override

class EllcieHAR(HARDataset):
    @dataclass
    class Row:
        T: str
        Ax: str
        Ay: str
        Az: str
        Gx: str
        Gy: str
        Gz: str
        P: str
        CLASS: str


    activitylist: Final[dict[str, Activities]] = {
        'STANDING': Activities.STANDING,
        'STAND_TO_SIT': Activities.STAND_TO_SIT,
        'SITTING': Activities.SITTING,
        'SIT_TO_STAND': Activities.SIT_TO_STAND,
        'WALKING': Activities.WALKING,
        'SIT_TO_LIE': Activities.SIT_TO_LIE,
        'LYING': Activities.LYING,
        'LIE_TO_SIT': Activities.LIE_TO_SIT,
        'WALKING_UPSTAIRS': Activities.WALKING_UPSTAIRS,
        'WALKING_DOWNSTAIRS': Activities.WALKING_DOWNSTAIRS,
        'RUNNING': Activities.RUNNING,
        'DRINKING': Activities.DRINKING,
        'DRIVING': Activities.DRIVING,
    }

    def __init__(self,
                 path: str = '',
                 variant: str = '1',
                 files: list[str] | None = None) -> None:
        super().__init__()
        self.__dr = DirectoryReader()
        self.__cr: CSVReader[EllcieHAR.Row] = CSVReader()
        self.__path = Path(path)
        self.__variant = variant
        self.__files = files

        if self.__variant not in ['PACK-2', 'UCA-EHAR']:
            raise ValueError('Unsupported variant ' + variant)

        self.sets.remove('valid')

    @override
    def __call__(self) -> HARDataModel:
        files = self.__dr.read(self.__path, ext='.csv')
        filesd = [self.__cr.read(f, delimiter=';', labels=EllcieHAR.Row) for f in files]

        subjects: dict[str, Subject] = {}

        for fd in filesd:
            name = fd.filename
            if self.__files is not None and not any(s.lower() in name.name.lower() for s in self.__files):
                continue

            data = fd.content
            subject_name = name.stem
            if self.__variant == 'PACK-2':
                subject_name = subject_name.split('_')[-2]
            elif self.__variant == 'UCA-EHAR':
                subject_name = subject_name.split('_')[1]

            if subject_name not in subjects:
                subjects[subject_name] = Subject(subject_name, [])
            activities = subjects.setdefault(subject_name, Subject(subject_name, [])).activities

            lastact = data[0].CLASS
            firstactt = float(data[0].T)
            samples: list[TimeSample] = []

            for d in data:
                if lastact != d.CLASS:
                    if lastact != '': # Drop unlabelled samples
                        activities.append(Activity(self.activitylist[lastact], samples))
                    samples = []
                    lastact = d.CLASS
                    firstactt = float(d.T)

                absolutet = float(d.T)

                sensors: list[Sensor] = []
                if d.Ax != '': # acc/gyro should be synced
                    sensors.append(Accelerometer(float(d.Ax), float(d.Ay), float(d.Az)))
                    sensors.append(Gyroscope(float(d.Gx), float(d.Gy), float(d.Gz)))
                if d.P != '':
                    sensors.append(Barometer(float(d.P)))
                sample = TimeSample(t=absolutet - firstactt, sensors=sensors)
                samples.append(sample)

            if d.CLASS != '': # Drop unlabelled samples
                activities.append(Activity(self.activitylist[d.CLASS], samples))

        return HARDataModel(sets=HARDataModel.Sets(train=list(subjects.values())), name=self.name)

    @property
    @override
    def name(self) -> str:
        return f'{super().name}_{self.__variant}'



from unittest import TestCase

class TestModelsCNN1D(TestCase):
    def setUp(self):
        import numpy as np
        from qualia_core.datamodel import RawDataModel
        from qualia_core.datamodel.RawDataModel import RawData, RawDataSets
        train = RawData(np.ones((2, 64, 1), dtype=np.float32), np.array([[1, 0], [1, 0]]))
        test = RawData(np.ones((2, 64, 1), dtype=np.float32), np.array([[1, 0], [0, 1]]))
        self.__data = RawDataModel(RawDataSets(train=train, test=test), name='test_cnn1d')
        self.__model_params = {
            'filters': (4, 6),
            'fc_units': (10, 10),
            'kernel_sizes': (5, 3),
            'batch_norm': True,
            'dropouts': [0.5, 0.5, 0, 0],
            'pool_sizes': (4, 2),
            'prepool': 2,
            'dims': 1
            }

    def test_cnn_1d_keras(self):
        from qualia_core import qualia
        from qualia_core.learningframework import Keras
        from qualia_core.learningmodel.keras import CNN
        from tensorflow.keras.layers import Dense, Activation, Flatten, Conv1D, AveragePooling1D, MaxPooling1D, BatchNormalization, Dropout
        from tensorflow.keras.activations import relu, softmax

        model = CNN

        framework = Keras()

        trainresult = qualia.train(self.__data,
                        train_epochs=1,
                        iteration=1,
                        model_name='test_cnn_1d_keras',
                        model=model,
                        model_params=self.__model_params,
                        optimizer={'kind': 'Adam'},
                        framework=framework,
                        )
        self.assertEqual(trainresult.name, 'test_cnn_1d_keras')
        self.assertEqual(trainresult.i, 1)
        self.assertEqual(trainresult.model.input_shape, (None, 64, 1))
        self.assertEqual(trainresult.model.output_shape, (None, 2))


        # prepool
        self.assertIsInstance(trainresult.model.layers[0], AveragePooling1D)
        self.assertEqual(tuple(trainresult.model.layers[0].pool_size), (2,))
        self.assertEqual(tuple(trainresult.model.layers[0].strides), (2,))

        # 1st conv block
        self.assertIsInstance(trainresult.model.layers[1], Conv1D)
        self.assertEqual(tuple(trainresult.model.layers[1].kernel.shape), (5, 1, 4)) # 4*1*5 = 20
        self.assertEqual(tuple(trainresult.model.layers[1].bias.shape), (4, )) # 4
        self.assertIsInstance(trainresult.model.layers[2], Activation)
        self.assertEqual(trainresult.model.layers[2].activation, relu)
        self.assertIsInstance(trainresult.model.layers[3], BatchNormalization)
        self.assertEqual(tuple(trainresult.model.layers[3].moving_mean.shape), (4,)) # 4
        self.assertEqual(tuple(trainresult.model.layers[3].moving_variance.shape), (4, )) # 4
        self.assertEqual(tuple(trainresult.model.layers[3].gamma.shape), (4, )) # 4
        self.assertEqual(tuple(trainresult.model.layers[3].beta.shape), (4, )) # 4
        self.assertIsInstance(trainresult.model.layers[4], Dropout)
        self.assertEqual(trainresult.model.layers[4].rate, 0.5)
        self.assertIsInstance(trainresult.model.layers[5], MaxPooling1D)
        self.assertEqual(tuple(trainresult.model.layers[5].pool_size), (4,))
        self.assertEqual(tuple(trainresult.model.layers[5].strides), (4,))

        # 2nd conv block
        self.assertIsInstance(trainresult.model.layers[6], Conv1D)
        self.assertEqual(tuple(trainresult.model.layers[6].kernel.shape), (3, 4, 6)) # 6*4*3 = 72
        self.assertEqual(tuple(trainresult.model.layers[6].bias.shape), (6, )) # 6
        self.assertIsInstance(trainresult.model.layers[7], Activation)
        self.assertEqual(trainresult.model.layers[7].activation, relu)
        self.assertIsInstance(trainresult.model.layers[8], BatchNormalization)
        self.assertEqual(tuple(trainresult.model.layers[8].moving_mean.shape), (6,)) # 6
        self.assertEqual(tuple(trainresult.model.layers[8].moving_variance.shape), (6, )) # 6
        self.assertEqual(tuple(trainresult.model.layers[8].gamma.shape), (6, )) # 6
        self.assertEqual(tuple(trainresult.model.layers[8].beta.shape), (6, )) # 6
        self.assertIsInstance(trainresult.model.layers[9], Dropout)
        self.assertEqual(trainresult.model.layers[9].rate, 0.5)
        self.assertIsInstance(trainresult.model.layers[10], MaxPooling1D)
        self.assertEqual(tuple(trainresult.model.layers[10].pool_size), (2,))
        self.assertEqual(tuple(trainresult.model.layers[10].strides), (2,))

        self.assertIsInstance(trainresult.model.layers[11], Flatten)

        # fc
        self.assertIsInstance(trainresult.model.layers[12], Dense)
        self.assertEqual(trainresult.model.layers[12].kernel.shape, (12, 10)) # 120
        self.assertEqual(trainresult.model.layers[12].bias.shape, (10, )) # 10
        self.assertIsInstance(trainresult.model.layers[13], Activation)
        self.assertEqual(trainresult.model.layers[13].activation, relu)
        self.assertIsInstance(trainresult.model.layers[14], Dense)
        self.assertEqual(trainresult.model.layers[14].kernel.shape, (10, 10)) # 100
        self.assertEqual(trainresult.model.layers[14].bias.shape, (10, )) # 10
        self.assertIsInstance(trainresult.model.layers[15], Activation)
        self.assertEqual(trainresult.model.layers[15].activation, relu)
        self.assertIsInstance(trainresult.model.layers[16], Dense)
        self.assertEqual(trainresult.model.layers[16].kernel.shape, (10, 2)) # 20
        self.assertEqual(trainresult.model.layers[16].bias.shape, (2, )) # 2

        # softmax
        self.assertIsInstance(trainresult.model.layers[17], Activation)
        self.assertEqual(trainresult.model.layers[17].activation, softmax)


        # first layer 10 weights/10 biases, second layer 10*10 weights (10 inputs, 10 outputs)/10 biases, 3rd layer 10*2 weights (10 inputs 2 outputs)/2 biases, 4 bytes (float32)
        self.assertEqual(trainresult.mem_params, (4*1*5 + 4 + 4*4 + 6*4*3 + 6 + 6*4 + 10*12 +10 + 10*10 + 10 + 2*10 + 2) * 4)
        self.assertEqual(trainresult.acc, 0.5) # Same data in one or the other class, should have 50% acc
        self.assertEqual(trainresult.framework, framework)

    def test_cnn_1d_pytorch(self):
        from qualia_core import qualia
        from qualia_core.learningframework import PyTorch
        from qualia_core.learningmodel.pytorch import CNN
        from torch.nn import Flatten, Linear, ReLU, AvgPool1d, MaxPool1d, Conv1d, BatchNorm1d, Dropout

        model = CNN

        framework = PyTorch(enable_progress_bar=False)

        trainresult = qualia.train(self.__data,
                        train_epochs=1,
                        iteration=1,
                        model_name='test_cnn_1d_pytorch',
                        model=model,
                        model_params=self.__model_params,
                        optimizer={'kind': 'Adam'},
                        framework=framework,
                        )
        self.assertEqual(trainresult.name, 'test_cnn_1d_pytorch')
        self.assertEqual(trainresult.i, 1)
        #self.assertEqual(trainresult.model.input_shape, (None, 1, 1)) # Not supported in PyTorch
        #self.assertEqual(trainresult.model.output_shape, (None, 2)), Not supported in PyTorch


        # prepool
        self.assertIsInstance(trainresult.model[0], AvgPool1d)
        self.assertEqual(tuple(trainresult.model[0].kernel_size), (2,))
        self.assertEqual(tuple(trainresult.model[0].stride), (2,))

        # 1st conv block
        self.assertIsInstance(trainresult.model[1], Conv1d)
        self.assertEqual(tuple(trainresult.model[1].weight.shape), (4, 1, 5)) # 4*1*5 = 20
        self.assertEqual(tuple(trainresult.model[1].bias.shape), (4, )) # 4
        self.assertIsInstance(trainresult.model[2], BatchNorm1d)
        self.assertEqual(tuple(trainresult.model[2].weight.shape), (4,)) # 4
        self.assertEqual(tuple(trainresult.model[2].bias.shape), (4, )) # 4
        self.assertIsInstance(trainresult.model[3], ReLU)
        self.assertIsInstance(trainresult.model[4], Dropout)
        self.assertEqual(trainresult.model[4].p, 0.5)
        self.assertIsInstance(trainresult.model[5], MaxPool1d)
        self.assertEqual(trainresult.model[5].kernel_size, 4)
        self.assertEqual(trainresult.model[5].stride, 4)

        # 2nd conv block
        self.assertIsInstance(trainresult.model[6], Conv1d)
        self.assertEqual(tuple(trainresult.model[6].weight.shape), (6, 4, 3)) # 6*4*3 = 72
        self.assertEqual(tuple(trainresult.model[6].bias.shape), (6, )) # 6
        self.assertIsInstance(trainresult.model[7], BatchNorm1d)
        self.assertEqual(tuple(trainresult.model[7].weight.shape), (6,)) # 6
        self.assertEqual(tuple(trainresult.model[7].bias.shape), (6, )) # 6
        self.assertIsInstance(trainresult.model[8], ReLU)
        self.assertIsInstance(trainresult.model[9], Dropout)
        self.assertEqual(trainresult.model[9].p, 0.5)
        self.assertIsInstance(trainresult.model[10], MaxPool1d)
        self.assertEqual(trainresult.model[10].kernel_size, 2)
        self.assertEqual(trainresult.model[10].stride, 2)

        self.assertIsInstance(trainresult.model[11], Flatten)

        # fc
        self.assertIsInstance(trainresult.model[12], Linear)
        self.assertEqual(trainresult.model[12].weight.shape, (10, 12)) # 120
        self.assertEqual(trainresult.model[12].bias.shape, (10, )) # 10
        self.assertIsInstance(trainresult.model[13], ReLU)
        self.assertIsInstance(trainresult.model[14], Linear)
        self.assertEqual(trainresult.model[14].weight.shape, (10, 10)) # 100
        self.assertEqual(trainresult.model[14].bias.shape, (10, )) # 10
        self.assertIsInstance(trainresult.model[15], ReLU)
        self.assertIsInstance(trainresult.model[16], Linear)
        self.assertEqual(trainresult.model[16].weight.shape, (2, 10)) # 20
        self.assertEqual(trainresult.model[16].bias.shape, (2, )) # 2


        self.assertEqual(trainresult.mem_params, (4*1*5 + 4 + 4*2 + 6*4*3 + 6 + 6*2 + 10*12 +10 + 10*10 + 10 + 2*10 + 2) * 4)
        self.assertEqual(trainresult.acc, 0.5) # Same data in one or the other class, should have 50% acc
        self.assertEqual(trainresult.framework, framework)



import sys

import torch

from .CustomBatchNorm import CustomBatchNorm

if sys.version_info >= (3, 12):
    from typing import override
else:
    from typing_extensions import override

class CustomBatchNorm2d(CustomBatchNorm):
    @override
    def forward(self, input: torch.Tensor) -> torch.Tensor:
        """Reshape to 1D-like input in order to use CustomBatchNorm forward as-is.

        BatchNorm works on channels with are second dim after batch so flatten the last dimensions to single dim.
        """
        input_shape = input.shape
        input = input.flatten(start_dim=2)

        y = super().forward(input)

        return y.reshape(input_shape)



from enum import auto

from qualia_core.utils import StrEnum

from enum import IntEnum

class Activities(IntEnum): #(StrEnum):
    NULL = auto()
    STANDING = auto()
    STAND_TO_SIT = auto()
    SITTING = auto()
    SIT_TO_STAND = auto()
    WALKING = auto()
    STOPPING = auto()
    LYING = auto()
    STAND_TO_LIE = auto()
    LIE_TO_STAND = auto()
    STAND_TO_SQUAT = auto()
    SQUAT_TO_STAND = auto()
    SQUATTING = auto()
    WALKING_DOWNSTAIRS = auto()
    WALKING_UPSTAIRS = auto()

    # Used in RealLife_HAR
    INACTIVE = auto()
    ACTIVE = auto()
    DRIVING = auto()

    # Used in EllcieHAR
    RUNNING = auto()
    DRINKING = auto()
    SIT_TO_LIE = auto()
    LIE_TO_SIT = auto()



from __future__ import annotations

import logging
import sys
from typing import Any, Callable

import numpy as np
import numpy.typing

from qualia_core.datamodel import RawDataModel
from qualia_core.datamodel.har.HARDataModel import HARDataModel
from qualia_core.datamodel.har.InfoHAR import Info, Info_dtype
from qualia_core.datamodel.RawDataModel import RawData, RawDataSets
from qualia_core.typing import TYPE_CHECKING

from .Preprocessing import Preprocessing

if TYPE_CHECKING:
    from qualia_core.datamodel.har.Activities import Activities  # noqa: TCH001
    from qualia_core.dataset.Dataset import Dataset  # noqa: TCH001

if sys.version_info >= (3, 12):
    from typing import override
else:
    from typing_extensions import override

logger = logging.getLogger(__name__)

class DatamodelConverter(Preprocessing[HARDataModel, RawDataModel]):
    @override
    def __call__(self, datamodel: HARDataModel) -> RawDataModel:
        # First generate the class numbers according to the activities present in the dataset
        activitylabels_set: set[Activities] = set()
        for _, s in datamodel.sets:
            for subject in s:
                for activity in subject.activities:
                    activitylabels_set.add(activity.kind)
        activitylabels = sorted(activitylabels_set)
        logger.info('Activity labels: %s', activitylabels)

        sets: dict[str, RawData] = {}
        for name, s in datamodel:
            data_x: list[list[numpy.typing.NDArray[np.float32]]] = []
            data_y: list[int] = []
            infos: list[Info] = []
            for subject in s:
                for i, activity in enumerate(subject.activities):
                    data_x += activity.get_raw_array()
                    #dataY += [int(activity.kind)] * len(activity.samples)
                    data_y += [activitylabels.index(activity.kind)] * len(activity.samples)
                    infos += [Info(subject.name, activity.kind, i)] * len(activity.samples)
            sets[name] = RawData(x=np.array(data_x),
                                 y=np.array(data_y),
                                 info=np.array(infos, dtype=Info_dtype))
        return RawDataModel(sets=RawDataSets(**sets), name=datamodel.name)

    @override
    def import_data(self, dataset: Dataset[Any]) -> Callable[[], RawDataModel | None]:
        return lambda: RawDataModel.import_data(name=dataset.name, sets=dataset.sets)



from pathlib import Path
import re
import sys
import time
from qualia_core.evaluation.Stats import Stats
import numpy as np
import multiprocessing

def benchmark(tag, model_kind, dataset, repeat, batch_size, use_gpu, ret):
    from qualia_core.utils import TensorFlowInitializer
    from qualia_core.learningframework import Keras

    tfi = TensorFlowInitializer()
    tfi(reserve_gpu=use_gpu, gpu_memory_growth=False)

    import qualia_core.learningmodel.keras as models
    model_kind = getattr(models, model_kind)

    framework = Keras()
    model = framework.load(tag, model=model_kind, path=Path('out')/'deploy'/'keras')

    print('Started')
    tiledset = np.tile(dataset.sets.test.x, (repeat, 1, 1))
    tstart = time.perf_counter_ns() # Start timer
    train_predictions = model.predict(tiledset, batch_size=batch_size)
    tstop = time.perf_counter_ns() # Stop timer
    avg_time = (tstop - tstart) / (1000000000 * len(tiledset))
    print('Finished')

    ret.value = avg_time

class Keras:
    def __init__(self, batch_size, repeat, use_gpu, *args, **kwargs):
        self.use_gpu = use_gpu
        self.batch_size = batch_size
        self.repeat = repeat


    def evaluate(self, model_kind, dataset, target: str, tag: str, limit: int=None):
        import tensorflow as tf
        from qualia_core.learningframework import Keras

        framework = Keras()
        model = framework.load(tag, model=model_kind, path=Path('out')/'deploy'/'keras')
        model.compile(metrics=['CategoricalAccuracy'])
        
        results = model.evaluate(dataset.sets.test.x, dataset.sets.test.y, batch_size=self.batch_size, return_dict=True)

        ctx = multiprocessing.get_context('spawn')
        ret = ctx.Value("d", 0.0, lock=False)
        p = ctx.Process(target=benchmark, args=(tag, model_kind, dataset, self.repeat, self.batch_size, self.use_gpu, ret))
        p.start()
        p.join()
        avg_time = ret.value

        train_predictions = model.predict(dataset.sets.test.x)
        print(tf.math.confusion_matrix(dataset.sets.test.y.argmax(axis=1), train_predictions.argmax(axis=1)))

        return Stats(avg_time=avg_time, accuracy=results['categorical_accuracy'])



import math
from collections import OrderedDict
from typing import Union

import numpy as np
from torch import nn


class CNN(nn.Sequential):
    def __init__(self,
        input_shape,
        output_shape,

        filters: list=[6, 16, 120],
        kernel_sizes: list=[3, 3, 5],
        paddings: list=[0, 0, 0],
        strides: list=[1, 1, 1],
        batch_norm: bool=False,
        dropouts: Union[float, list[float]] = 0.0,
        pool_sizes: list=[2, 2, 0],
        fc_units: list=[84],
        prepool: int=1,
        postpool: int=1,

        gsp: bool=False,

        dims: int=1):

        self.input_shape = input_shape
        self.output_shape = output_shape

        if dims == 1:
            import qualia_core.learningmodel.pytorch.layers1d as layers_t
        elif dims == 2:
            import qualia_core.learningmodel.pytorch.layers2d as layers_t
        else:
            raise ValueError('Only dims=1 or dims=2 supported')

        # Backward compatibility for config not defining dropout as a list
        if not isinstance(dropouts, list):
            dropouts = [dropouts] * (len(filters) + len(fc_units))

        layers = OrderedDict()

        if not isinstance(prepool, int) and math.prod(prepool) > 1 or prepool > 1:
            layers['prepool'] = layers_t.AvgPool(prepool)

        layers['conv1'] = layers_t.Conv(in_channels=input_shape[-1], out_channels=filters[0], kernel_size=kernel_sizes[0], padding=paddings[0], stride=strides[0])

        if batch_norm:
            layers['bn1'] = layers_t.BatchNorm(filters[0])

        layers['relu1'] = nn.ReLU()

        if dropouts[0]:
            layers['dropout1'] = nn.Dropout(dropouts[0])
        if pool_sizes[0]:
            layers['maxpool1'] = layers_t.MaxPool(pool_sizes[0])

        i = 2
        for in_filters, out_filters, kernel, pool_size, padding, stride, dropout in zip(filters, filters[1:], kernel_sizes[1:],
                                                                                        pool_sizes[1:], paddings[1:], strides[1:],
                                                                                        dropouts[1:]):
            layers[f'conv{i}'] = layers_t.Conv(in_channels=in_filters, out_channels=out_filters, kernel_size=kernel, padding=padding, stride=stride)

            if batch_norm:
                layers[f'bn{i}'] = layers_t.BatchNorm(out_filters)

            layers[f'relu{i}'] = nn.ReLU()

            if dropout:
                layers[f'dropout{i}'] = nn.Dropout(dropout)
            if pool_size:
                layers[f'maxpool{i}'] = layers_t.MaxPool(pool_size)

            i += 1

        if not isinstance(postpool, int) and math.prod(postpool) > 1 or postpool > 1:
            layers['postpool'] = layers_t.AvgPool(postpool)

        if gsp:
            layers['gsp'] = layers_t.GlobalSumPool()
        else:
            layers['flatten'] = nn.Flatten()

            in_features = np.array(input_shape[:-1]) // np.array(prepool)
            for _, kernel, pool, padding, stride in zip(filters, kernel_sizes, pool_sizes, paddings, strides):
                in_features += np.array(padding) * 2
                in_features -= (kernel - 1)
                in_features = np.ceil(in_features / stride).astype(int)
                if pool:
                    in_features = in_features // pool
            in_features = in_features // np.array(postpool)
            in_features = in_features.prod()
            in_features *= filters[-1]

            j = 1
            for in_units, out_units, dropout in zip((in_features, *fc_units), fc_units, dropouts[len(filters):]):
                layers[f'fc{j}'] = nn.Linear(in_units, out_units)
                layers[f'relu{i}'] = nn.ReLU()
                if dropout:
                    layers[f'dropout{i}'] = nn.Dropout(dropout)
                i += 1
                j += 1

            layers[f'fc{j}'] = nn.Linear(fc_units[-1] if len(fc_units) > 0 else in_features, output_shape[0])

        super().__init__(layers)



from __future__ import annotations

import sys

from qualia_core.datamodel.har.HARDataModel import HARDataModel
from qualia_core.datamodel.har.Subject import Subject

from .Dataset import Dataset

if sys.version_info >= (3, 12):
    from typing import override
else:
    from typing_extensions import override

class HARDataset(Dataset[list[Subject]]):
    def __init__(self) -> None:
        super().__init__(sets=list(HARDataModel.Sets.fieldnames()))

    @override
    def import_data(self) -> HARDataModel | None:
        raise NotImplementedError



from .DeployAndEvaluate import DeployAndEvaluate
from .ParameterResearch import ParameterResearch
from .PrepareDeploy import PrepareDeploy
from .PreprocessData import PreprocessData
from .Train import Train

__all__ = ['DeployAndEvaluate', 'ParameterResearch', 'PrepareDeploy', 'PreprocessData', 'Train']



import logging
import time

import numpy as np

from .Preprocessing import Preprocessing

logger = logging.getLogger(__name__)

class MFCC(Preprocessing):
    def __init__(self,
                 sample_rate: int,
                 n_mfcc: int,
                 dct_type: int=2,
                 norm: str='ortho',
                 log_mels=False,
                 melkwargs: dict={},
                 chunks: int = 4,
                 dims: int = 1):
        import torchaudio
        self.mfcc_transform = torchaudio.transforms.MFCC(sample_rate=sample_rate,
                                                        n_mfcc=n_mfcc,
                                                        dct_type=dct_type,
                                                        norm=norm,
                                                        log_mels=log_mels,
                                                        melkwargs=melkwargs)
        self.__chunks = chunks
        self.__dims = dims

    def __call__(self, datamodel):
        start = time.time()
        import torch
        for name, s in datamodel:
            # 1D
            s.x = s.x.squeeze(-1)
            s.x = torch.tensor(s.x)
            chunks = torch.chunk(s.x, self.__chunks)
            chunks_mfcc = [self.mfcc_transform(chunk) for chunk in chunks]
            s.x = torch.concat(chunks_mfcc)
            s.x = s.x.numpy()
            s.x = s.x.swapaxes(1, 2)
            if self.__dims == 2:
                s.x = np.expand_dims(s.x, -1)

            # 2D, add channels dim
            #s.x = self.mfcc_transform(torch.tensor(s.x.squeeze(-1))).unsqueeze(-1).numpy()

        logger.info('Elapsed: %s s', time.time() - start)
        return datamodel



from __future__ import annotations

import copy
from datetime import datetime
from pathlib import Path
from typing import Any, NamedTuple

import colorful as cf  # type: ignore[import-untyped]

from qualia_core.typing import TYPE_CHECKING
from qualia_core.utils.merge_dict import merge_dict

if TYPE_CHECKING:
    from qualia_core.dataaugmentation.DataAugmentation import DataAugmentation  # noqa: TCH001
    from qualia_core.datamodel.RawDataModel import RawDataModel  # noqa: TCH001
    from qualia_core.learningframework.LearningFramework import LearningFramework  # noqa: TCH001
    from qualia_core.typing import ConfigDict
    from qualia_core.utils import Git  # noqa: TCH001
    from qualia_core.utils.logger import Logger  # noqa: TCH001
    from qualia_core.utils.plugin import QualiaComponent  # noqa: TCH001

class ParameterResearchLoggerFields(NamedTuple):
    name: str
    params: int
    mem_params: int
    accuracy: float
    tune_config: str

class ParameterResearch:
    def __call__(self,
                 qualia: QualiaComponent,
                 learningframework: LearningFramework[Any],
                 dataaugmentations: list[DataAugmentation],
                 data: RawDataModel,
                 config: ConfigDict,
                 git: Git) -> dict[str, Logger[ParameterResearchLoggerFields]]:
        import optuna

        from qualia_core.qualia import train
        from qualia_core.utils.logger import CSVLogger

        loggers: dict[str, CSVLogger[ParameterResearchLoggerFields]] = {}

        log: CSVLogger[ParameterResearchLoggerFields] = CSVLogger('parameter_research',
                                                                  file=Path('parameter_research')/f'{git.short_hash()}_{datetime.now():%Y-%m-%d_%H-%M-%S}.csv')
        loggers['parameter_research'] = log
        # Write column names
        log.fields = ParameterResearchLoggerFields

        (Path('out')/'learningmodel').mkdir(parents=True, exist_ok=True)

        experimenttracking = config.get('experimenttracking', None)

        if config['bench']['first_run'] != config['bench']['last_run']:
            print('Warning: parameter_research does not follow first_run/last_run')

        # Filter out disabled models
        config['model'] = [m for m in config['model'] if not m.get('disabled', False)]

        if isinstance(config['model'], list):
            if len(config['model']) > 1:
                raise ValueError('parameter_research only runs on a single model definition')
            else:
                config['model'] = config['model'][0]
        if config['model'].get('disabled', False):
            raise ValueError('model must not be disabled for parameter_research')


        def training_function(trial):
            def gen_tune_config(params): # In-place
                if isinstance(params, list):
                    for i, v in enumerate(params):
                        if 'kind' in v and 'params' in v: # "leaf" to call trial function on
                            params[i] = getattr(trial, v['kind'])(**v['params'])
                        else:
                            gen_tune_config(i)
                else:
                    print(params)
                    for k, v in params.items():
                        if 'kind' in v and 'params' in v: # "leaf" to call trial function on
                            params[k] = getattr(trial, v['kind'])(**v['params'])
                        else:
                            gen_tune_config(v)
                return params

            print(f'{cf.bold}Hyperparameter research for {cf.blue}{config["model"]["name"]}{cf.close_fg_color}, trial {cf.red}{trial.number}{cf.reset}')

            tune_config = gen_tune_config(copy.deepcopy(config['parameter_research']['params']['trial']))

            print(f'{cf.bold}Trial params:{cf.reset} {tune_config=}')

            m = copy.deepcopy(tune_config)
            m = merge_dict(m, config['model'])

            print(f'{cf.bold}Model params:{cf.reset} {m=}')

            et = None
            if experimenttracking is not None:
                et = getattr(learningframework.experimenttrackings, experimenttracking['kind'])(**experimenttracking.get('params', {}))
                et.hyperparameters = {'config': config, 'model': m, 'i': trial.number}


            trainresult = train(datamodel=data,
                    train_epochs=m.get('epochs', None),
                    iteration=trial.number,
                    model_name=m["name"],
                    model=getattr(learningframework.learningmodels, m['kind']),
                    model_params=m.get('params', {}),
                    batch_size=m.get('batch_size', None),
                    optimizer=m.get('optimizer', {}),
                    framework=learningframework,
                    load=m.get('load', False),
                    train=m.get('train', True),
                    evaluate=m.get('evaluate', True),
                    dataaugmentations=dataaugmentations,
                    experimenttracking=et)

            if et is not None:
                et.stop()

            log(ParameterResearchLoggerFields(name=trainresult.name,
                                              params=trainresult.params,
                                              mem_params=trainresult.mem_params,
                                              accuracy=trainresult.acc,
                                              tune_config=str(tune_config)))

            for postprocessing in config.get('postprocessing', []):
                ppp = {k: v for k,v in postprocessing.get('params', {}).items()} # Workaround tomlkit bug where some nested dict would lose their items
                trainresult = getattr(qualia.postprocessing, postprocessing['kind'])(**ppp)(
                    trainresult=trainresult,
                    model_conf=m)
                if trainresult.log:
                    log(ParameterResearchLoggerFields(name=trainresult.name,
                                                      params=trainresult.params,
                                                      mem_params=trainresult.mem_params,
                                                      accuracy=trainresult.acc,
                                                      tune_config=str(tune_config)))
                if postprocessing.get('export', False):
                    trainresult.framework.export(trainresult.model, f'{trainresult.name}_r{i}')

            return trainresult.acc



        if config['parameter_research'].get('params', {}).get('study', {}).get('load', False):
            study = optuna.load_study(**config['parameter_research'].get('params', {}).get('study', {}).get('params', {}))
        else:
            study = optuna.create_study(**config['parameter_research'].get('params', {}).get('study', {}).get('params', {}))
        study.optimize(training_function, **config['parameter_research'].get('params', {}).get('optimize', {}).get('params', {}))

        return loggers



from pathlib import Path
from typing import Callable

import dill
from filelock import FileLock
import pytest


def uci_har_preprocess_data() -> dict:
    import qualia_core.utils.config
    from qualia_core import main
    configf, configname = qualia_core.utils.config.parse_config(Path('conf')/'tests'/'UCI-HAR_ResNetv1_train_float32.toml')
    config = qualia_core.utils.config.validate_config_dict(configf)
    return main.qualia('preprocess_data', config, configname)

def uci_har_resnet_train_float32(fixture_uci_har_preprocess_data: Callable[[], dict]) -> dict:
    from qualia_core import main
    import qualia_core.utils.config
    configf, configname = qualia_core.utils.config.parse_config(Path('conf')/'tests'/'UCI-HAR_ResNetv1_train_float32.toml')
    config = qualia_core.utils.config.validate_config_dict(configf)
    return {**main.qualia('train', config, configname), **fixture_uci_har_preprocess_data}

def uci_har_resnet_train_int16(fixture_uci_har_preprocess_data: Callable[[], dict],
                               fixture_uci_har_resnet_train_float32: Callable[[], dict]) -> dict:
    from qualia_core import main
    import qualia_core.utils.config
    configf, configname = qualia_core.utils.config.parse_config(Path('conf')/'tests'/'UCI-HAR_ResNetv1_train_int16.toml')
    config = qualia_core.utils.config.validate_config_dict(configf)
    return {**main.qualia('train', config, configname), **fixture_uci_har_preprocess_data}

def uci_har_resnet_train_int8(fixture_uci_har_preprocess_data: Callable[[], dict],
                              fixture_uci_har_resnet_train_float32: Callable[[], dict]) -> dict:
    from qualia_core import main
    import qualia_core.utils.config
    configf, configname = qualia_core.utils.config.parse_config(Path('conf')/'tests'/'UCI-HAR_ResNetv1_train_int8.toml')
    config = qualia_core.utils.config.validate_config_dict(configf)
    return {**main.qualia('train', config, configname), **fixture_uci_har_preprocess_data}

def xdist_shared(worker_id: str, tmp_path_factory, name: str, callback: Callable[[], dict]):
    if worker_id == 'master':
        return callback()

    root_tmp_dir = tmp_path_factory.getbasetemp().parent

    fn = root_tmp_dir / f'{name}.json'
    with FileLock(str(fn) + '.lock'):
        if fn.is_file():
            from qualia_core.utils.TensorFlowInitializer import TensorFlowInitializer
            # workaround to prevent deserialization from loading a keras/TF module without calling TensorFlowInitializer
            TensorFlowInitializer()(seed=2, reserve_gpu=False, gpu_memory_growth=True)

            with fn.open('rb') as f:
                data = dill.load(f)
        else:
            data = callback()
            with fn.open('wb') as f:
                dill.dump(data, f, protocol=dill.HIGHEST_PROTOCOL)
    return data

@pytest.fixture(scope='package')
def fixture_uci_har_preprocess_data(worker_id: str, tmp_path_factory) -> dict:
    return xdist_shared(worker_id, tmp_path_factory, 'fixture_uci_har_preprocess_data', uci_har_preprocess_data)

@pytest.fixture(scope='package')
def fixture_uci_har_resnet_train_float32(worker_id: str,
                                         tmp_path_factory,
                                         fixture_uci_har_preprocess_data: Callable[[], dict]) -> dict:
    return xdist_shared(worker_id, tmp_path_factory, 'fixture_uci_har_resnet_train_float32', lambda: uci_har_resnet_train_float32(fixture_uci_har_preprocess_data))


@pytest.fixture(scope='package')
def fixture_uci_har_resnet_train_int16(worker_id: str,
                                       tmp_path_factory,
                                       fixture_uci_har_preprocess_data: Callable[[], dict],
                                       fixture_uci_har_resnet_train_float32: Callable[[], dict]) -> dict:
    return xdist_shared(worker_id, tmp_path_factory, 'fixture_uci_har_resnet_train_int16', lambda:
                        uci_har_resnet_train_int16(fixture_uci_har_preprocess_data, fixture_uci_har_resnet_train_float32))

@pytest.fixture(scope='package')
def fixture_uci_har_resnet_train_int8(worker_id: str,
                                      tmp_path_factory,
                                      fixture_uci_har_preprocess_data: Callable[[], dict],
                                      fixture_uci_har_resnet_train_float32: Callable[[], dict]) -> dict:
    return xdist_shared(worker_id, tmp_path_factory, 'fixture_uci_har_resnet_train_int8', lambda:
                        uci_har_resnet_train_int8(fixture_uci_har_preprocess_data, fixture_uci_har_resnet_train_float32))



from __future__ import annotations

import logging
import sys
from typing import NamedTuple, overload

from qualia_core.typing import TYPE_CHECKING

from .Logger import Logger
from .TextFormatter import TextFormatter

if TYPE_CHECKING:
    from pathlib import Path  # noqa: TCH003

if sys.version_info >= (3, 12):
    from typing import override
else:
    from typing_extensions import override

error_logger = logging.getLogger(__name__)

class TextLoggerField(NamedTuple):
    content: str

class TextLogger(Logger[TextLoggerField]):
    def __init__(self,
                 name: str,
                 file: Path | None = None) -> None:
        super().__init__(name=name, file=file, suffix='.txt', fields=TextLoggerField, log_fields=False, formatter=TextFormatter())

    @overload
    def __call__(self, data: str) -> None:
        ...

    @overload
    def __call__(self, data: TextLoggerField) -> None:
        ...

    @override
    def __call__(self, data: str | TextLoggerField) -> None:
        if self.logger is None:
            self._lazy_init()
        if self.logger is None:
            error_logger.error('Could not initialize logger %s', self._name)
            raise RuntimeError

        if not isinstance(data, TextLoggerField):
            self.logger.info(str(data))
            self._content.append(TextLoggerField(content=data))
        else:
            self.logger.info(data.content)
            self._content.append(data)



import numpy as np

from .Preprocessing import Preprocessing

class Reshape2DTo1D(Preprocessing):
    def __call__(self, datamodel):
        import torch
        for name, s in datamodel:
            s.x = s.x.reshape((s.x.shape[0], s.x.shape[1] * s.x.shape[2], s.x.shape[3]))
            print(s.x.shape)

        return datamodel



from pathlib import Path
import re
import sys
from qualia_core.utils.process import subprocesstee                                                                                     
from qualia_core.evaluation.Stats import Stats

import qualia_core.evaluation.STM32CubeAI as STM32CubeAIBase

class STM32CubeAI(STM32CubeAIBase):
    def __init__(self, *args, **kwargs):
        super().__init__(mode='x86')



from __future__ import annotations

import sys

from qualia_core.typing import TYPE_CHECKING

from .DataAugmentationPyTorch import DataAugmentationPyTorch

if TYPE_CHECKING:
    import torch  # noqa: TCH002

if sys.version_info >= (3, 12):
    from typing import override
else:
    from typing_extensions import override


class ResizedCrop(DataAugmentationPyTorch):
    def __init__(self,  # noqa: PLR0913
                 size: list[int],
                 scale: list[float] | None = None,
                 ratio: list[float] | None = None,
                 antialias: bool = True,  # noqa: FBT001, FBT002
                 evaluate: bool = False,  # noqa: FBT001, FBT002
                 before: bool = False,  # noqa: FBT001, FBT002
                 after: bool = True) -> None:  # noqa: FBT001, FBT002
        import torchvision.transforms  # type: ignore[import-untyped]

        super().__init__(evaluate=evaluate, before=before, after=after)

        scale = scale if scale is not None else [0.08, 1.0]
        ratio = ratio if ratio is not None else [0.75, 4/3]
        self.randomresizedcrop = torchvision.transforms.RandomResizedCrop(size=size,
                                                                          scale=tuple(scale),
                                                                          ratio=tuple(ratio),
                                                                          antialias=antialias)

    def apply(self, x: torch.Tensor, y: torch.Tensor) -> tuple[torch.Tensor, torch.Tensor]:
        resizedcropped_x = self.randomresizedcrop(x)

        return resizedcropped_x, y

    @override
    def __call__(self, data: tuple[torch.Tensor, torch.Tensor], device: torch.device) -> tuple[torch.Tensor, torch.Tensor]:
        return self.apply(*data)



from .AutoAugment import AutoAugment
from .Crop import Crop
from .GaussianNoise import GaussianNoise
from .HorizontalFlip import HorizontalFlip
from .IntToFloat32 import IntToFloat32
from .MFCC import MFCC
from .Mixup import Mixup
from .ResizedCrop import ResizedCrop
from .Rotation import Rotation
from .Rotation2D import Rotation2D
from .TimeShifting import TimeShifting
from .TimeWarping import TimeWarping

__all__ = ['AutoAugment', 'Crop', 'GaussianNoise', 'HorizontalFlip', 'IntToFloat32', 'MFCC', 'Mixup', 'ResizedCrop', 'Rotation', 'Rotation2D', 'TimeShifting', 'TimeWarping']



from __future__ import annotations

import logging
import os
import sys
import time
import wave
from concurrent.futures import Future, ProcessPoolExecutor
from multiprocessing.managers import SharedMemoryManager
from multiprocessing.shared_memory import SharedMemory
from pathlib import Path
from typing import Any, Callable, Final

import numpy as np
import numpy.typing
from qualia_core.datamodel import RawDataModel
from qualia_core.datamodel.RawDataModel import RawData, RawDataSets
from qualia_core.utils.file import DirectoryReader

from .RawDataset import RawDataset

if sys.version_info >= (3, 12):
    from typing import override
else:
    from typing_extensions import override

logger = logging.getLogger(__name__)

class GSC(RawDataset):
    class_list_no_background_noise: Final[dict[str, int | None]] = {
        '_background_noise_': None, # Drop background noise
        'backward': 0,
        'bed': 1,
        'bird': 2,
        'cat': 3,
        'dog': 4,
        'down': 5,
        'eight': 6,
        'five': 7,
        'follow': 8,
        'forward': 9,
        'four': 10,
        'go': 11,
        'happy': 12,
        'house': 13,
        'learn': 14,
        'left': 15,
        'marvin': 16,
        'nine': 17,
        'no': 18,
        'off': 19,
        'on': 20,
        'one': 21,
        'right': 22,
        'seven': 23,
        'sheila': 24,
        'six': 25,
        'stop': 26,
        'three': 27,
        'tree': 28,
        'two': 29,
        'up': 30,
        'visual': 31,
        'wow': 32,
        'yes': 33,
        'zero': 34,
    }

    class_list_digits: Final[dict[str, int | None]] = {
        '_background_noise_': None, # Drop background noise
        'backward': None,
        'bed': None,
        'bird': None,
        'cat': None,
        'dog': None,
        'down': None,
        'eight': 8,
        'five': 5,
        'follow': None,
        'forward': None,
        'four': 4,
        'go': None,
        'happy': None,
        'house': None,
        'learn': None,
        'left': None,
        'marvin': None,
        'nine': 9,
        'no': None,
        'off': None,
        'on': None,
        'one': 1,
        'right': None,
        'seven': 7,
        'sheila': None,
        'six': 6,
        'stop': None,
        'three': 3,
        'tree': None,
        'two': 2,
        'up': None,
        'visual': None,
        'wow': None,
        'yes': None,
        'zero': 0,
    }

    def __init__(self, path: str, variant: str = 'v2', subset: str = 'digits', train_valid_split: bool = False) -> None:
        super().__init__()
        self.__path = Path(path)
        self.__variant = variant
        self.__subset = subset
        self.__dr = DirectoryReader()
        self.__train_valid_split = train_valid_split
        if not train_valid_split:
            self.sets.remove('valid')

    def load_wave(self, recording: Path) -> numpy.typing.NDArray[np.float32]:
        with wave.open(str(recording)) as f:
            data = f.readframes(f.getnframes())

        data_array = np.frombuffer(data, dtype=np.int16).copy().astype(np.float32)
        data_array.resize((16000, 1)) # Resize to 1s (16kHz) with zero-padding, 1 channel
        return data_array

    def _load_files(self,
                    i: int,
                    files: numpy.typing.NDArray[Any], # Path or object not supported as NDArray generic
                    loader: Callable[[Path],
                                     numpy.typing.NDArray[np.float32]]) -> tuple[str, tuple[int, ...], numpy.typing.DTypeLike]:
        start = time.time()

        data_list: list[numpy.typing.NDArray[np.float32]] = []
        logger.info('Process %s loading %s files...', i, len(files))

        for file in files:
            data = loader(file)
            data_list.append(data)

        data_array = np.array(data_list)
        del data_list

        data_buffer = SharedMemory(size=data_array.nbytes, create=True)

        data_shared = np.frombuffer(data_buffer.buf, dtype=data_array.dtype).reshape(data_array.shape)

        np.copyto(data_shared, data_array)

        del data_shared

        ret = (data_buffer.name, data_array.shape, data_array.dtype)

        data_buffer.close()

        logger.info('Process %s finished in %s s.', i, time.time() - start)
        return ret

    def __threaded_loader(self,
                          training_files: list[Path],
                          testing_files: list[Path],
                          validation_files: list[Path],
                          loader: Callable[[Path],
                                           numpy.typing.NDArray[np.float32]]) -> tuple[numpy.typing.NDArray[np.float32] | None,
                                                                                       numpy.typing.NDArray[np.float32] | None,
                                                                                       numpy.typing.NDArray[np.float32] | None]:
        cpus: int | None = os.cpu_count()
        total_chunks: int = cpus // 2 if cpus is not None else 2
        total_files = len(training_files) + len(validation_files) + len(testing_files)
        training_chunks = min(len(training_files), max(1, len(training_files) * total_chunks // total_files))
        validation_chunks = min(len(validation_files), max(1, len(validation_files) * total_chunks // total_files))
        testing_chunks = min(len(testing_files), max(1, total_chunks - training_chunks - validation_chunks))
        training_files_chunks = np.array_split(np.array(training_files), training_chunks) if training_files else []
        validation_files_chunks = np.array_split(np.array(validation_files), validation_chunks) if validation_files else []
        testing_files_chunks = np.array_split(np.array(testing_files), testing_chunks) if testing_files else []

        logger.info('Using %s threads for training data, %s threads for validation data and %s threads for testing data',
                    training_chunks,
                    validation_chunks,
                    testing_chunks)

        with SharedMemoryManager() as smm:
            with ProcessPoolExecutor() as executor:
                train_futures = [executor.submit(self._load_files, i, files, loader)
                           for i, files in enumerate(training_files_chunks)]
                valid_futures = [executor.submit(self._load_files, i, files, loader)
                           for i, files in enumerate(validation_files_chunks)]
                test_futures = [executor.submit(self._load_files, i, files, loader)
                           for i, files in enumerate(testing_files_chunks)]

                def load_results(futures: list[Future[tuple[str,
                                                            tuple[int, ...],
                                                            numpy.typing.DTypeLike]]]) -> numpy.typing.NDArray[np.float32] | None:

                    names = [f.result()[0] for f in futures]
                    shapes = [f.result()[1] for f in futures]
                    dtypes = [f.result()[2] for f in futures]
                    bufs = [SharedMemory(n) for n in names]

                    data_list = [np.frombuffer(buf.buf, dtype=dtype).reshape(shape)
                              for shape, dtype, buf in zip(shapes, dtypes, bufs)]

                    data_array = np.concatenate(data_list) if data_list else None
                    del data_list

                    for buf in bufs:
                        buf.unlink()

                    return data_array

                train_x_array = load_results(train_futures)
                valid_x_array = load_results(valid_futures)
                test_x_array = load_results(test_futures)

        return train_x_array, valid_x_array, test_x_array


    def __load_v2(self, path: Path, class_list: dict[str, int | None]) -> RawDataModel:
        start = time.time()

        directory = self.__dr.read(path, ext='.wav', recursive=True)

        with (path/'validation_list.txt').open() as f:
            validation_list = f.read().splitlines()

        with (path/'testing_list.txt').open() as f:
            testing_list = f.read().splitlines()

        # Build files list for train and test
        training_files: list[Path] = []
        validation_files: list[Path] = []
        testing_files: list[Path] = []
        training_labels: list[int] = []
        validation_labels: list[int] = []
        testing_labels: list[int] = []
        for file in list(directory):
            label = class_list[file.parent.name]
            if label is None: # Drop sample excluded from class list
                continue
            if str(file.relative_to(path)) in testing_list:
                testing_files.append(file)
                testing_labels.append(label)
            elif self.__train_valid_split and str(file.relative_to(path)) in validation_list:
                validation_files.append(file)
                validation_labels.append(label)
            else:
                training_files.append(file)
                training_labels.append(label)

        train_x, valid_x, test_x = self.__threaded_loader(training_files, testing_files, validation_files, loader=self.load_wave)

        train_y = np.array(training_labels) if training_labels else None
        valid_y = np.array(validation_labels) if validation_labels else None
        test_y = np.array(testing_labels) if testing_labels else None
        logger.info('Shapes: train_x=%s, train_y=%s, valid_x=%s, valid_y=%s, test_x=%s, test_y=%s',
                    train_x.shape if train_x is not None else None,
                    train_y.shape if train_y is not None else None,
                    valid_x.shape if valid_x is not None else None,
                    valid_y.shape if valid_y is not None else None,
                    test_x.shape if test_x is not None else None,
                    test_y.shape if test_y is not None else None)

        train = RawData(train_x, train_y) if train_x is not None and train_y is not None else None
        valid = RawData(valid_x, valid_y) if valid_x is not None and valid_y is not None else None
        test = RawData(test_x, test_y) if test_x is not None and test_y is not None else None

        logger.info('Elapsed: %s s', time.time() - start)

        return RawDataModel(sets=RawDataSets(train=train, valid=valid, test=test), name=self.name)

    @override
    def __call__(self) -> RawDataModel:
        if self.__variant != 'v2':
            raise ValueError('Only v2 variant supported')

        if self.__subset == 'digits':
            class_list = GSC.class_list_digits
        elif self.__subset == 'no_background_noise':
            class_list = GSC.class_list_no_background_noise
        else:
            raise ValueError('Only digits or no_background_noise subsets supported')

        return self.__load_v2(self.__path, class_list=class_list)

    @property
    @override
    def name(self) -> str:
        return f'{super().name}_{self.__variant}_{self.__subset}'



import numpy as np

from .Preprocessing import Preprocessing

class Window(Preprocessing):
    '''Warning: Must be applied before Class2BinMatrix and DatasetSplitter'''
    def __init__(self, size: int=0, stride: int=1, no_overlapping_labels: bool=True, unique_label_per_window: bool=True):
        self.__size = size
        self.__stride = stride
        self.__no_overlapping_labels = no_overlapping_labels
        self.__unique_label_per_window = unique_label_per_window

    # Inspired from https://stackoverflow.com/a/45730836 with support added for multidimensional arrays, windowing in 1st dim
    def __window(self, a, size=4, stride=2):
        shape = (a.shape[0] - size + 1, size) + a.shape[1:]
        strides = a.strides[:1] * 2 + a.strides[1:]
        view = np.lib.stride_tricks.as_strided(a, strides=strides, shape=shape)
        view = view[0::stride]
        return view.copy()

    def __call__(self, datamodel):
        for name, s in datamodel:
            if len(s.data) < 1: # Ignore empty sets
                continue
            s.data = self.__window(s.data, size=self.__size, stride=self.__stride)
            s.labels = self.__window(s.labels, size=self.__size, stride=self.__stride)
            s.info = self.__window(s.info, size=self.__size, stride=self.__stride)

            if self.__no_overlapping_labels:
                is_not_overlapping = np.all(s.labels == s.labels[:,0].reshape((-1, 1)), axis=-1)
                non_overlapping_i = np.where(is_not_overlapping)

                s.data = s.data[non_overlapping_i]
                s.labels = s.labels[non_overlapping_i]
                s.info = s.info[non_overlapping_i]

            print(f'Number of vectors in {name} set: {len(s.data)}')

            if self.__unique_label_per_window:
                s.labels = np.array([np.bincount(w).argmax() for w in s.y])

                print(f'Number of vector per class in {name} set: {np.bincount(s.labels)}')
                print(f'Balancing weights: {np.max(np.bincount(s.labels)) / np.bincount(s.labels)}')

        return datamodel



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any, Generic, TypeVar

from qualia_core.typing import TYPE_CHECKING, OptimizerConfigDict

if TYPE_CHECKING:
    from types import ModuleType  # noqa: TCH003

    import numpy.typing  # noqa: TCH002

    from qualia_core.dataaugmentation.DataAugmentation import DataAugmentation  # noqa: TCH001
    from qualia_core.datamodel.RawDataModel import RawData  # noqa: TCH001
    from qualia_core.experimenttracking.ExperimentTracking import ExperimentTracking  # noqa: TCH001

T = TypeVar('T')

class LearningFramework(ABC, Generic[T]):
    learningmodels: ModuleType

    @abstractmethod
    def train(self,
              model: T,
              trainset: RawData,
              validationset: RawData,
              epochs: int,
              batch_size: int,
              optimizer: OptimizerConfigDict | None,
              dataaugmentations: list[DataAugmentation],
              experimenttracking: ExperimentTracking | None,
              name: str) -> T:
        pass

    @abstractmethod
    def load(self, name: str, model: T) -> T:
        pass

    @abstractmethod
    def evaluate(self,
                 model: T,
                 testset: RawData,
                 batch_size: int,
                 dataaugmentations: list[DataAugmentation],
                 experimenttracking: ExperimentTracking | None,
                 dataset_type: str,
                 name: str) -> dict[str, int | float | numpy.typing.NDArray[Any]]:
        pass

    @abstractmethod
    def predict(self,
                 model: T,
                 dataset: RawData,
                 batch_size: int,
                 dataaugmentations: list[DataAugmentation],
                 experimenttracking: ExperimentTracking | None,
                 name: str) -> Any:
        ...

    @abstractmethod
    def export(self, model: T, name: str) -> None:
        pass

    @abstractmethod
    def summary(self, model: T) -> None:
        pass

    @abstractmethod
    def n_params(self, model: T) -> int:
        pass

    @abstractmethod
    def save_graph_plot(self, model: T, model_save: str) -> None:
        pass

    @abstractmethod
    def apply_dataaugmentation(self,
                               da: DataAugmentation,
                               x: numpy.typing.NDArray[Any],
                               y: numpy.typing.NDArray[Any],
                               **kwargs: Any) -> tuple[numpy.typing.NDArray[Any], numpy.typing.NDArray[Any]]:
        ...



from __future__ import annotations

import logging
import sys
from pathlib import Path

import numpy as np

from qualia_core.datamodel import RawDataModel
from qualia_core.datamodel.RawDataModel import RawData, RawDataSets

from .RawDataset import RawDataset

if sys.version_info >= (3, 12):
    from typing import override
else:
    from typing_extensions import override

logger = logging.getLogger(__name__)

class WSMNIST(RawDataset):
    def __init__(self, path: str, variant: str='spoken') -> None:
        super().__init__()
        self.__path = Path(path)
        self.__variant = variant
        self.sets.remove('valid')

    def __load_spoken(self, path: Path) -> RawDataModel:
        train_x = np.load(path/'data_sp_train.npy').reshape(-1, 39, 13).astype(np.float32)
        train_y = np.load(path/'labels_train.npy')
        test_x = np.load(path/'data_sp_test.npy').reshape(-1, 39, 13).astype(np.float32)
        test_y = np.load(path/'labels_test.npy')

        train = RawData(train_x, train_y)
        test = RawData(test_x, test_y)

        return RawDataModel(sets=RawDataSets(train=train, test=test), name=self.name)

    def __load_written(self, path: Path) -> RawDataModel:
        train_x = np.load(path/'data_wr_train.npy').reshape(-1, 28, 28, 1).astype(np.float32)
        train_y = np.load(path/'labels_train.npy')
        test_x = np.load(path/'data_wr_test.npy').reshape(-1, 28, 28, 1).astype(np.float32)
        test_y = np.load(path/'labels_test.npy')

        train = RawData(train_x, train_y)
        test = RawData(test_x, test_y)

        return RawDataModel(sets=RawDataSets(train=train, test=test), name=self.name)

    @override
    def __call__(self) -> RawDataModel:
        if self.__variant == 'spoken':
            return self.__load_spoken(self.__path)
        if self.__variant == 'written':
            return self.__load_written(self.__path)

        logger.error('Only spoken or written variants supported')
        raise ValueError

    @property
    @override
    def name(self) -> str:
        return f'{super().name}_{self.__variant}'



from typing import ClassVar
from qualia_core.datamodel.RawDataModel import RawDataModel


class TestModelsMLP:
    data: ClassVar[RawDataModel]

    @classmethod
    def setup_method(cls) -> None:
        import numpy as np
        from qualia_core.datamodel import RawDataModel
        from qualia_core.datamodel.RawDataModel import RawData, RawDataSets
        train = RawData(np.array([[[1.0]], [[1.0]]], dtype=np.float32), np.array([[1, 0], [1, 0]]))
        test = RawData(np.array([[[1.0]], [[1.0]]], dtype=np.float32), np.array([[1, 0], [0, 1]]))
        cls.data = RawDataModel(RawDataSets(train=train, test=test), name='test_mlp')

    def test_mlp_keras(self):
        from qualia_core import qualia
        from qualia_core.learningframework import Keras
        from qualia_core.learningmodel.keras import MLP
        from tensorflow.keras.layers import Dense, Activation, Flatten
        from tensorflow.keras.activations import relu, softmax

        model = MLP
        model_params = {'units': (10, 10)}

        framework = Keras()

        trainresult = qualia.train(TestModelsMLP.data,
                        train_epochs=1,
                        iteration=1,
                        model_name='test_mlp_keras',
                        model=model,
                        model_params=model_params,
                        optimizer={'kind': 'Adam'},
                        framework=framework,
                        )
        assert trainresult.name == 'test_mlp_keras'
        assert trainresult.i == 1
        assert trainresult.model.input_shape == (None, 1, 1)
        assert trainresult.model.output_shape == (None, 2)

        assert isinstance(trainresult.model.layers[0], Flatten)
        assert isinstance(trainresult.model.layers[1], Dense)
        assert trainresult.model.layers[1].kernel.shape == (1, 10)
        assert trainresult.model.layers[1].bias.shape == (10)
        assert isinstance(trainresult.model.layers[2], Activation)
        assert trainresult.model.layers[2].activation == relu
        assert isinstance(trainresult.model.layers[3], Dense)
        assert trainresult.model.layers[3].kernel.shape == (10, 10)
        assert trainresult.model.layers[3].bias.shape == (10)
        assert isinstance(trainresult.model.layers[4], Activation)
        assert trainresult.model.layers[4].activation == relu
        assert isinstance(trainresult.model.layers[5], Dense)
        assert trainresult.model.layers[5].kernel.shape == (10, 2)
        assert trainresult.model.layers[5].bias.shape == (2)
        assert isinstance(trainresult.model.layers[6], Activation)
        assert trainresult.model.layers[6].activation == softmax

        # first layer 10 weights/10 biases, second layer 10*10 weights (10 inputs, 10 outputs)/10 biases, 3rd layer 10*2 weights (10 inputs 2 outputs)/2 biases, 4 bytes (float32)
        assert trainresult.mem_params == (10 + 10 + 10*10 + 10 + 10*2 + 2) * 4
        assert trainresult.acc == 0.5 # Same data in one or the other class, should have 50% acc
        assert trainresult.framework == framework

    def test_mlp_pytorch(self) -> None:
        from qualia_core import qualia
        from qualia_core.learningframework import PyTorch
        from qualia_core.learningmodel.pytorch import MLP
        from torch.nn import Flatten, Linear, ReLU

        model = MLP
        model_params = {'units': (10, 10)}

        framework = PyTorch(enable_progress_bar=False)

        trainresult = qualia.train(TestModelsMLP.data,
                        train_epochs=1,
                        iteration=1,
                        model_name='test_mlp_pytorch',
                        model=model,
                        model_params=model_params,
                        optimizer={'kind': 'Adam'},
                        framework=framework,
                        )
        assert trainresult.name == 'test_mlp_pytorch'
        assert trainresult.i == 1
        #self.assert trainresult.model.input_shape, (None, 1, 1)) # Not supported in PyTorch
        #self.assert trainresult.model.output_shape, (None, 2)), Not supported in PyTorch
        # first layer 10 weights/10 biases, second layer 10*10 weights (10 inputs, 10 outputs)/10 biases, 3rd layer 10*2 weights (10 inputs 2 outputs)/2 biases, 4 bytes (float32)

        assert len(list(trainresult.model.children())) == 1
        assert len(list(trainresult.model.layers.children())) == 6
        assert isinstance(trainresult.model.layers.flatten1, Flatten)
        assert isinstance(trainresult.model.layers.fc1, Linear)
        assert trainresult.model.layers.fc1.weight.shape == (10, 1)
        assert trainresult.model.layers.fc1.bias.shape == (10, )
        assert isinstance(trainresult.model.layers.relu1, ReLU)
        assert isinstance(trainresult.model.layers.fc2, Linear)
        assert trainresult.model.layers.fc2.weight.shape == (10, 10)
        assert trainresult.model.layers.fc2.bias.shape == (10, )
        assert isinstance(trainresult.model.layers.relu2, ReLU)
        assert isinstance(trainresult.model.layers.fc3, Linear)
        assert trainresult.model.layers.fc3.weight.shape == (2, 10)
        assert trainresult.model.layers.fc3.bias.shape == (2, )

        assert trainresult.mem_params == (10 + 10 + 10*10 + 10 + 10*2 + 2) * 4
        assert trainresult.acc == 0.5 # Same data in one or the other class, should have 50% acc
        assert trainresult.framework, framework



import os
import glob
from pathlib import Path

class DirectoryReader:
    def read(self, directory: Path=None, ext: str='', recursive: bool=False):
        """
        Return the list of files in a given directory, optionally matching a given extension.

        Remark: case-sensitive.

        :param directory: directory to search in, if empty search is performed in working directory
        :param ext: optional file extension to match, '.' has to be inserted by the caller as required
        :param recursive: do a recursive search
        """
        directory = directory.expanduser()
        return directory.glob(f'{"**/" if recursive else ""}*{ext}')



from pathlib import Path

import numpy as np

import matplotlib.pyplot as plt
from matplotlib.backends.backend_pdf import PdfPages

from .Preprocessing import Preprocessing

class VisualizeActivities(Preprocessing):
    def __init__(self, activities, noconvert: bool=False):
        from qualia_core.datamodel.har import Activities
        for i, activity in enumerate(activities):
            if not noconvert and isinstance(activity, str):
                activities[i] = Activities[activity]
        self.__activities = activities


    def __call__(self, datamodel):
        outpath = Path('out')/'visualization'/'activities'/datamodel.name

        fig = plt.figure()
        for sname, s in datamodel.sets:
            for subject in s:
                outdir = outpath/sname/subject.name
                outdir.mkdir(parents=True, exist_ok=True)

                to_plot = {}
                
                for i, activity in enumerate(subject.activities):
                    if activity.kind in self.__activities:
                        to_plot.setdefault(activity.kind, []).append([i, activity])

                for kind, acts in to_plot.items():

                    outfile = outdir/f'{kind.name}.pdf'

                    pp = PdfPages(outfile)

                    for (activity_i, act) in acts:
                        samples = np.array(act.get_raw_array())
                        samples = samples.reshape((samples.shape[0], -1))
                        for i in range(samples.shape[-1]):
                            p = plt.plot(samples[:,i])

                        plt.suptitle(activity_i)

                        pp.savefig(fig)
                        fig.clf()
                    pp.close()
        return datamodel



class LearningModel:
    input_shape: tuple[int, ...]
    output_shape: tuple[int, ...]

    def __init__(self, input_shape: tuple[int, ...], output_shape: tuple[int, ...]) -> None:
        super().__init__()
        self.input_shape = input_shape
        self.output_shape = output_shape



from .test_cnn1d import TestModelsCNN1D
from .test_cnn2d import TestModelsCNN2D
from .test_mlp import TestModelsMLP
from .test_quantizedresnet1d import TestModelsQuantizedResnet1D
from .test_quantizedresnet2d import TestModelsQuantizedResnet2D
from .test_resnet1d import TestModelResNet1D
from .test_resnet2d import TestModelsResNet2D



from pathlib import Path
import shutil
from collections import namedtuple
from qualia_core.utils.process import subprocesstee

class Linux:
    import qualia_core.evaluation.host.STM32CubeAI as evaluator # Suggested evaluator

    def prepare(self, tag, model, optimize: str, compression: int):
        raise NotImplemented('Should write TFLite model')
        print(self.__class__.__name__, 'Info: running locally, nothing to prepare')
        return self

    def deploy(self, tag):
        print(self.__class__.__name__, 'Info: running locally, nothing to deploy')

        return namedtuple('Deploy', ['rom_size', 'evaluator'])(self.__rom_size(tag), self.evaluator)

    def __rom_size(self, tag):
        return -1




from __future__ import annotations

import sys

from qualia_core.experimenttracking.NeptuneBase import NeptuneBase
from qualia_core.typing import TYPE_CHECKING, RecursiveConfigDict

from .ExperimentTrackingPyTorch import ExperimentTrackingPyTorch

if sys.version_info >= (3, 12):
    from typing import override
else:
    from typing_extensions import override

if TYPE_CHECKING:
    from neptune.new.integrations.pytorch_lightning import NeptuneLogger  # noqa: TCH002


class Neptune(NeptuneBase, ExperimentTrackingPyTorch):
    def __init__(self, project_name: str, config_file: str='conf/neptune.toml') -> None:
        super().__init__(project_name=project_name, config_file=config_file)

    @override
    def start(self, name: str | None = None) -> None:
        from neptune.new.integrations.pytorch_lightning import NeptuneLogger

        project_name = f'{self.project_name}_{name}' if name is not None else self.project_name

        self.neptune_logger = NeptuneLogger(
            api_key=self.api_key,
            project=f'{self.project_namespace}/{project_name}',
            source_files=self.source_files,
            close_after_fit=False)

    @override
    def stop(self) -> None:
        self.neptune_logger.experiment.stop()

    @override
    def _hyperparameters(self, params: RecursiveConfigDict) -> None:
        for k, v in params.items():
            self.neptune_logger.experiment[k].log(v)

    @property
    def logger(self) -> NeptuneLogger:
        return self.neptune_logger




import importlib.util
import sys

__all__ = []

if importlib.util.find_spec("tensorflow") is None:
    print('Warning: cannot find TensorFlow, Keras models will be unavailable', file=sys.stderr)
elif importlib.util.find_spec("tensorflow.keras") is None:
    print('Warning: cannot find TensorFlow.Keras, Keras models will be unavailable', file=sys.stderr)
else:
    from . import keras
    __all__ += ['keras']

if importlib.util.find_spec("torch") is None:
    print('Warning: cannot find PyTorch, PyTorch models will be unavailable', file=sys.stderr)
elif importlib.util.find_spec("pytorch_lightning") is None:
    print('Warning: cannot find PyTorch Lightning, PyTorch models will be unavailable', file=sys.stderr)
else:
    from . import pytorch
    __all__ += ['pytorch']




from .Activities import Activities
from .Activity import Activity

class TimedActivity(Activity):
    def __init__(self, kind: Activities, timesamples: 'dict[float, TimeSample]'= None):
        self.kind = kind
        self.timesamples = timesamples if timesamples is not None else {}

    @property
    def samples(self):
        keys = list(self.timesamples.keys())
        keys.sort()
        return [self.timesamples[k] for k in keys]

    @samples.setter
    def samples(self, newsamples):
        self.timesamples = {s.t: s for s in newsamples}



import sys

import torch
from torch import nn

from qualia_core.learningmodel.LearningModel import LearningModel

if sys.version_info >= (3, 12):
    from typing import override
else:
    from typing_extensions import override


class LearningModelPyTorch(LearningModel, nn.Module):
    def __init__(self, input_shape: tuple[int, ...], output_shape: tuple[int, ...]) -> None:
        super().__init__(input_shape=input_shape, output_shape=output_shape)

    @override
    def __call__(self, input: torch.Tensor) -> torch.Tensor:  # noqa: A002
        return super().__call__(input)



from __future__ import annotations

import sys

from qualia_core.datamodel.RawDataModel import RawData, RawDataModel

from .Dataset import Dataset

if sys.version_info >= (3, 12):
    from typing import override
else:
    from typing_extensions import override

class RawDataset(Dataset[RawData]):
    def __init__(self) -> None:
        super().__init__(sets=list(RawDataModel.Sets.fieldnames()))

    @override
    def import_data(self) -> RawDataModel | None:
        return RawDataModel.import_data(name=self.name, sets=self.sets)



from __future__ import annotations

import math
import sys
from collections import OrderedDict

from torch import nn

from qualia_core.learningmodel.pytorch.layers.quantized_layers import QuantizedIdentity, QuantizedLinear
from qualia_core.learningmodel.pytorch.LearningModelPyTorch import LearningModelPyTorch
from qualia_core.typing import TYPE_CHECKING

if TYPE_CHECKING:
    import torch

    from qualia_core.typing import QuantizationConfigDict

if sys.version_info >= (3, 12):
    from typing import override
else:
    from typing_extensions import override

class QuantizedMLP(LearningModelPyTorch):
    def __init__(self,
                 input_shape: tuple[int, ...],
                 output_shape: tuple[int, ...],
                 units: list[int],
                 quant_params: QuantizationConfigDict) -> None:
        super().__init__(input_shape=input_shape, output_shape=output_shape)

        self.input_shape = input_shape
        self.output_shape = output_shape

        layers: OrderedDict[str, nn.Module] = OrderedDict()

        layers['identity1'] = QuantizedIdentity(quant_params=quant_params)
        layers['flatten1'] = nn.Flatten()

        i = 1
        for in_units, out_units in zip([math.prod(input_shape), *units], units[:-1]):
            layers[f'fc{i}'] = QuantizedLinear(in_units,
                                          out_units,
                                          quant_params=quant_params,
                                          activation=nn.ReLU())
            i += 1

        layers[f'fc{i}'] = QuantizedLinear(units[-1] if len(units) > 1 else math.prod(input_shape),
                                      output_shape[0],
                                      quant_params=quant_params)

        self.layers = nn.ModuleDict(layers)

    @override
    def forward(self, input: torch.Tensor) -> torch.Tensor:  # noqa: A002
        """Forward calls each of the SCNN :attr:`layers` sequentially.

        :param input: Input tensor
        :return: Output tensor
        """
        x = input
        for layer in self.layers:
            x = self.layers[layer](x)

        return x



from .Activities import Activities
from .Activity import Activity
from .HARDataModel import HARDataModel
from .Subject import Subject
from .TimedActivity import TimedActivity

__all__ = ['Activities', 'Activity', 'HARDataModel', 'Subject', 'TimedActivity']



import sys

import torch
from torch import nn

if sys.version_info >= (3, 12):
    from typing import override
else:
    from typing_extensions import override

class GlobalSumPool1d(nn.Module):
    def __init__(self) -> None:
        self.call_super_init = True # Support multiple inheritance from nn.Module
        super().__init__()

    @override
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return x.sum(dim=(-1))



from __future__ import annotations

import copy
import importlib.util
import logging
import sys
from pathlib import Path
from typing import TYPE_CHECKING, Any, Callable, cast

import qualia_core.deployment.qualia_codegen

from .Converter import Converter

if sys.version_info >= (3, 12):
    from typing import override
else:
    from typing_extensions import override

if TYPE_CHECKING:
    from torch import nn  # noqa: I001 # torch must be imported before keras to avoid deadlock
    import keras  # type: ignore[import] # No stubs for keras package
    import numpy.typing
    from qualia_codegen_core.graph import ModelGraph
    from qualia_codegen_core.graph.layers import TBaseLayer

    from qualia_core.learningframework.LearningFramework import LearningFramework

logger = logging.getLogger(__name__)

class QualiaCodeGen(Converter[Any]):
    deployers = qualia_core.deployment.qualia_codegen #: Suggested deployers

    __number_type: type[int | float]
    _h: str | None = None
    _name: str | None = None

    def __init__(self,
                 quantize: str,
                 long_width: int | None = None,
                 outdir: str | None = None) -> None:
        super().__init__()

        self.__quantize = quantize
        self.__outdir = Path(outdir) if outdir is not None else Path('out')/'qualia_codegen'

        if quantize == 'float32':
            self.__number_type = float
            self.__width = 32
            self.__long_width = 32 if long_width is None else long_width
        elif quantize == 'int16':
            self.__number_type = int
            self.__width = 16
            self.__long_width = 32 if long_width is None else long_width
        elif quantize == 'int8':
            self.__number_type = int
            self.__width = 8
            self.__long_width = 16 if long_width is None else long_width
        else:
            logger.error('Qualia-CodeGen only supports no (float32) quantization, int8 or int16 quantization, got %s', quantize)
            raise ValueError

    def convert_model_to_modelgraph(self, model: nn.Module | keras.Model) -> ModelGraph | None:
        from qualia_codegen_core.graph.layers import TAddLayer, TSumLayer

        modelgraph: ModelGraph | None = None

        if importlib.util.find_spec('torch') is None:
            logger.warning('Cannot find PyTorch, PyTorch support for Qualia-CodeGen will be unavailable')
        else:
            from torch import nn
            if isinstance(model, nn.Module):
                from qualia_codegen_core.graph import TorchModelGraph

                from qualia_core.learningmodel.pytorch.layers import Add, GlobalSumPool1d, GlobalSumPool2d
                custom_layers: dict[type[nn.Module], Callable[[nn.Module, TBaseLayer], tuple[type[TBaseLayer], list[Any]]]] = {
                        Add: lambda *_: (TAddLayer, []),
                        GlobalSumPool1d: lambda *_: (TSumLayer, [(-1,)]),
                        GlobalSumPool2d: lambda *_: (TSumLayer, [(-2, -1)]),
                        }
                modelgraph = TorchModelGraph(model).convert(custom_layers=custom_layers)

        if importlib.util.find_spec('keras') is None:
            logger.warning('Cannot find Keras, Keras support for Qualia-CodeGen will be unavailable')
        else:
            import keras  # type: ignore[import] # No stubs for keras package
            if isinstance(model, keras.Model):
                from qualia_codegen_core.graph import KerasModelGraph
                modelgraph = KerasModelGraph(model).convert()

        return modelgraph

    def convert_modelgraph_to_c(self, modelgraph: ModelGraph, output_path: Path) -> str | None:
        from qualia_codegen_core import Converter
        converter = Converter(output_path=output_path)
        return converter.convert_model(modelgraph)

    @override
    def convert(self,
                framework: LearningFramework[nn.Module | keras.Model],
                model: nn.Module | keras.Model,
                model_name: str,
                representative_dataset: numpy.typing.NDArray[Any]) -> QualiaCodeGen | None:
        from qualia_codegen_core.graph import Quantization
        from qualia_codegen_core.graph.ActivationsRange import ActivationsRange
        from qualia_codegen_core.graph.RoundMode import RoundMode

        self._name = f'{model_name}_q{self.__quantize}'

        framework.summary(model)

        modelgraph = self.convert_model_to_modelgraph(model)
        if modelgraph is None:
            logger.error('Could not convert model to ModelGraph')
            return None

        if self.__number_type == int: # Activation range only when using fixed-point quantization
            activations_range = ActivationsRange()

            if importlib.util.find_spec('keras') is not None:
                from qualia_codegen_core.graph import KerasModelGraph
                if isinstance(modelgraph, KerasModelGraph):
                    activations_range = activations_range.load(
                                            Path('out')/'learningmodel'/f'{model_name}_activations_range.h5.txt',
                                            cast(str, model.layers[0].name))

            if importlib.util.find_spec('torch') is not None:
                from qualia_codegen_core.graph import TorchModelGraph
                if isinstance(modelgraph, TorchModelGraph):
                    activations_range = activations_range.load(
                                            Path('out')/'learningmodel'/f'{model_name}_activations_range.txt',
                                            'input')

            # Populate quantization information for all layers from activations_range
            for node in modelgraph.nodes:
                if node.layer.name in activations_range:
                    node.q = Quantization(
                            number_type=self.__number_type,
                            width=self.__width,
                            long_width=self.__long_width,
                            weights_scale_factor=activations_range[node.layer.name].weights_q,
                            output_scale_factor=activations_range[node.layer.name].activation_q,
                            weights_round_mode=activations_range[node.layer.name].weights_round_mode,
                            output_round_mode=activations_range[node.layer.name].activation_round_mode,
                            )
                else:
                    if not node.innodes:
                        logger.error('No quantization information for %s, and no previous layer to copy from.', node.layer.name)
                        return self
                    logger.warning('No quantization information for %s, applying first previous layer %s information',
                                   node.layer.name,
                                   node.innodes[0].layer.name)
                    node.q = copy.deepcopy(node.innodes[0].q)
        else:
            for node in modelgraph.nodes:
                # No scale factor if not fixed-point quantization on integers
                node.q = Quantization(
                        number_type=self.__number_type,
                        width=self.__width,
                        long_width=self.__long_width,
                        weights_scale_factor=0,
                        output_scale_factor=0,
                        weights_round_mode=RoundMode.NONE,
                        output_round_mode=RoundMode.NONE,
                        )
        # self.directory cannot be None as long as we define self._name above
        self._h = self.convert_modelgraph_to_c(modelgraph, output_path=cast(Path, self.directory))

        if self._h is None:
            logger.error('Could not convert ModelGraph to C')
            return None

        with (self.__outdir/self._name/'full_model.h').open('w') as f:
            _ = f.write(self._h)

        return self

    @property
    def h(self) -> str | None:
        return self._h

    @property
    def name(self) -> str | None:
        return self._name

    @property
    def directory(self) -> Path | None:
        if self.name is None:
            return None
        return self.__outdir / self.name

    @override
    def process_mem_params(self, mem_params: int) -> Callable[[LearningFramework[nn.Module | keras.Model],
                                                               nn.Module | keras.Model],
                                                              int]:
        def f(framework: LearningFramework[nn.Module | keras.Model], model: nn.Module | keras.Model) -> int:
            return (framework.n_params(model) * self.__width) // 8
        return f



from __future__ import annotations

import sys

from qualia_core.typing import TYPE_CHECKING, RecursiveConfigDict

from .DataAugmentationPyTorch import DataAugmentationPyTorch

if TYPE_CHECKING:
    import torch  # noqa: TCH002

if sys.version_info >= (3, 12):
    from typing import override
else:
    from typing_extensions import override

class MFCC(DataAugmentationPyTorch):
    def __init__(self,  # noqa: PLR0913
                 sample_rate: int,
                 n_mfcc: int,
                 dct_type: int = 2,
                 norm: str = 'ortho',
                 log_mels: bool = False,  # noqa: FBT002, FBT001
                 melkwargs: RecursiveConfigDict | None = None,
                 evaluate: bool = False,  # noqa: FBT001, FBT002
                 before: bool = False,  # noqa: FBT001, FBT002
                 after: bool = True) -> None:  # noqa: FBT001, FBT002
        import torchaudio  # type: ignore[import-untyped]

        super().__init__(evaluate=evaluate, before=before, after=after)

        self.mfcc_transform = torchaudio.transforms.MFCC(sample_rate=sample_rate,
                                                        n_mfcc=n_mfcc,
                                                        dct_type=dct_type,
                                                        norm=norm,
                                                        log_mels=log_mels,
                                                        melkwargs=melkwargs)

    def apply(self, x: torch.Tensor, y: torch.Tensor, device: torch.device) -> tuple[torch.Tensor, torch.Tensor]:
        return self.mfcc_transform.to(device)(x.squeeze(1)), y # Only handle single channel

    @override
    def __call__(self, data: tuple[torch.Tensor, torch.Tensor], device: torch.device) -> tuple[torch.Tensor, torch.Tensor]:
        return self.apply(*data, device=device)



from torch.nn import Conv2d as Conv
from torch.nn import BatchNorm2d as BatchNorm
from torch.nn import AvgPool2d as AvgPool
from torch.nn import AdaptiveAvgPool2d as AdaptiveAvgPool
from torch.nn import MaxPool2d as MaxPool
from .quantized_layers2d import QuantizedBatchNorm2d as QuantizedBatchNorm
from .quantized_layers2d import QuantizedConv2d as QuantizedConv
from .quantized_layers2d import QuantizedMaxPool2d as QuantizedMaxPool
from .quantized_layers2d import QuantizedAvgPool2d as QuantizedAvgPool
from .quantized_layers2d import QuantizedAdaptiveAvgPool2d as QuantizedAdaptiveAvgPool
from .layers import GlobalSumPool2d as GlobalSumPool
from .layers import QuantizedGlobalSumPool2d as QuantizedGlobalSumPool

__all__ = ['Conv',
           'BatchNorm',
           'AvgPool',
           'AdaptiveAvgPool',
           'MaxPool',
           'QuantizedBatchNorm',
           'QuantizedConv',
           'QuantizedMaxPool',
           'QuantizedAvgPool',
           'QuantizedAdaptiveAvgPool',
           'GlobalSumPool',
           'QuantizedGlobalSumPool']



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, Callable, Generic, TypeVar

if TYPE_CHECKING:
    from qualia_core.dataset.Dataset import Dataset

T = TypeVar('T')
U = TypeVar('U')

class Preprocessing(ABC, Generic[T, U]):
    @abstractmethod
    def __call__(self, datamodel: T) -> U:
        ...

    def import_data(self, dataset: Dataset[Any]) -> Dataset[Any] | Callable[[], U | None]:
        """no-op if the preprocessing doesn't modify the way of importing the dataset."""
        return dataset



from tensorflow.keras.layers import Conv2D as Conv
from tensorflow.keras.layers import AveragePooling2D as AveragePooling
from tensorflow.keras.layers import MaxPooling2D as MaxPooling
from tensorflow.keras.layers import ZeroPadding2D as ZeroPadding



from pathlib import Path

import tomlkit

from .ExperimentTracking import ExperimentTracking


class NeptuneBase(ExperimentTracking):
    project_name: str
    api_key: str
    project_namespace: str
    source_files: list[str]

    def __init__(self, project_name: str, config_file: str='conf/neptune.toml') -> None:
        super().__init__()

        self.project_name = project_name

        with Path(config_file).open() as f:
            config_dict = tomlkit.parse(f.read())

            self.api_key = config_dict['neptune']['api_key']
            self.project_namespace = config_dict['neptune']['project_namespace']
            self.source_files = config_dict['neptune']['source_files']



from unittest import TestCase

class TestTrainingFrameworks(TestCase):
    def setUp(self):
        import numpy as np
        from qualia_core.datamodel import RawDataModel
        from qualia_core.datamodel.RawDataModel import RawData, RawDataSets
        train = RawData(np.array([[[1.0]], [[1.0]]], dtype=np.float32), np.array([[1, 0], [1, 0]]))
        test = RawData(np.array([[[1.0]], [[1.0]]], dtype=np.float32), np.array([[1, 0], [0, 1]]))
        self.__data = RawDataModel(RawDataSets(train=train, test=test), name='test_train')

    def test_train_keras(self):
        from qualia_core import qualia
        from qualia_core.learningframework import Keras
        from tensorflow.keras.layers import Input, Flatten, Dense, Activation
        from tensorflow.keras.models import Sequential

        model = lambda input_shape, *a, **kwa: Sequential((Input(input_shape), Flatten(), Dense(2), Activation('softmax')))

        framework = Keras()

        trainresult = qualia.train(self.__data,
                        train_epochs=1,
                        iteration=1,
                        model_name='test_train',
                        model=model,
                        optimizer={'kind': 'Adam'},
                        framework=framework,
                        )
        self.assertEqual(trainresult.name, 'test_train')
        self.assertEqual(trainresult.i, 1)
        self.assertEqual(trainresult.model.input_shape, (None, 1, 1))
        self.assertEqual(trainresult.model.output_shape, (None, 2))
        self.assertEqual(trainresult.mem_params, (2 + 2) * 4) # 2 weights, 2 biases, 4 bytes (float32)
        self.assertEqual(trainresult.acc, 0.5) # Same data in one or the other class, should have 50% acc
        self.assertEqual(trainresult.framework, framework)

    def test_train_pytorch(self):
        from qualia_core import qualia
        from qualia_core.learningframework import PyTorch
        import math
        from torch.nn import Sequential, Flatten, Linear, ReLU

        class Model(Sequential):
            def __init__(self, input_shape, output_shape):
                self.input_shape = input_shape
                self.output_shape = output_shape
                super().__init__(Flatten(), Linear(math.prod(input_shape), 2))

        framework = PyTorch(enable_progress_bar=False)

        trainresult = qualia.train(self.__data,
                        train_epochs=1,
                        iteration=1,
                        model_name='test_train',
                        model=Model,
                        optimizer={'kind': 'Adam'},
                        framework=framework,
                        )
        self.assertEqual(trainresult.name, 'test_train')
        self.assertEqual(trainresult.i, 1)
        #self.assertEqual(trainresult.model.input_shape, (None, 1, 1)) # Not supported in PyTorch
        #self.assertEqual(trainresult.model.output_shape, (None, 2)), Not supported in PyTorch
        self.assertEqual(trainresult.mem_params, (2 + 2) * 4) # 2 weights, 2 biases, 4 bytes (float32)
        self.assertEqual(trainresult.acc, 0.5) # Same data in one or the other class, should have 50% acc
        self.assertEqual(trainresult.framework, framework)



from .Preprocessing import Preprocessing

class Normalize(Preprocessing):
    def z_score(self, data):
        if self.debug:
            print(f'Before normalization: train mean={data.sets.train.x.mean()}, train std={data.sets.train.x.std()}, test mean={data.sets.test.x.mean()}, test std={data.sets.test.x.std()}')

        x_mean = data.sets.train.x.mean(axis=self.__axis, keepdims=True)
        x_std = data.sets.train.x.std(axis=self.__axis, keepdims=True)

        for _, s in data:
            s.x -= x_mean
            s.x /= x_std

        if self.debug:
            print(f'After normalization: train mean={data.sets.train.x.mean()}, train std={data.sets.train.x.std()}, test mean={data.sets.test.x.mean()}, test std={data.sets.test.x.std()}')

        return data

    def min_max(self, data):
        if self.debug:
            print(f'Before normalization: train mean={data.sets.train.x.mean()}, train std={data.sets.train.x.std()}, test mean={data.sets.test.x.mean()}, test std={data.sets.test.x.std()}')
            print(f'                      train min={data.sets.train.x.min()}, train max={data.sets.train.x.max()}, test min={data.sets.test.x.min()}, test max={data.sets.test.x.max()}')

        x_min = data.sets.train.x.min(axis=tuple(self.__axis), keepdims=True)
        x_max = data.sets.train.x.max(axis=tuple(self.__axis), keepdims=True)

        for _, s in data:
            s.x -= x_min
            s.x /= (x_max - x_min)

        if self.debug:
            print(f'After normalization: train mean={data.sets.train.x.mean()}, train std={data.sets.train.x.std()}, test mean={data.sets.test.x.mean()}, test std={data.sets.test.x.std()}')
            print(f'                     train min={data.sets.train.x.min()}, train max={data.sets.train.x.max()}, test min={data.sets.test.x.min()}, test max={data.sets.test.x.max()}')

        return data

    methods = {
        'z-score': z_score,
        'min-max': min_max
    }

    def __init__(self, method: str='z-score', axis: tuple = (0,), debug: bool=False):
        self.__refdata = None
        self.debug = debug
        if method not in self.methods:
            raise ValueError(f'{self.__class__.__name__}: Method {method} is not supported. Supported methods: {", ".join(self.methods)}')
        else:
            self.__method = self.methods[method].__get__(self)
        self.__axis = axis

    def __call__(self, *args, **kwargs):
        return self.__method(*args, **kwargs)



from __future__ import annotations

import logging
import shutil
import sys
from abc import abstractmethod
from typing import Any

from qualia_core.deployment.Deploy import Deploy
from qualia_core.deployment.Deployer import Deployer
from qualia_core.typing import TYPE_CHECKING
from qualia_core.utils.process import subprocesstee

if TYPE_CHECKING:
    if sys.version_info >= (3, 11):
        from typing import Self
    else:
        from typing_extensions import Self

    from pathlib import Path  # noqa: TCH003

    from qualia_core.postprocessing.Converter import Converter  # noqa: TCH001

if sys.version_info >= (3, 12):
    from typing import override
else:
    from typing_extensions import override

logger = logging.getLogger(__name__)

class Eclipse(Deployer):

    def __init__(self,
        eclipse_bin: Path,
        size_bin: Path,
        upload_bin: Path,
        projectname: str,
        projectdir: Path,
        outdir: Path,
        buildtype: str = 'Release') -> None:
        super().__init__()

        self._projectname = projectname
        self._buildtype = buildtype
        self._workspacedir = outdir/'workspace'
        self._projectdir = projectdir

        self._outdir = outdir

        self.__eclipse_bin = eclipse_bin
        self.__size_bin = size_bin
        self.__upload_bin = upload_bin

    def _run(self,
              cmd: str | Path,
              *args: str,
              cwd: Path | None = None,
              env: dict[str, str] | None = None) -> bool:
        logger.info('Running: %s %s', cmd, ' '.join(args))
        returncode, _ = subprocesstee.run(str(cmd), *args, cwd=cwd, env=env)
        return returncode == 0

    def _create_outdir(self) -> None:
        self._outdir.mkdir(parents=True, exist_ok=True)
        self._workspacedir.mkdir(parents=True, exist_ok=True)

    def _clean_workspace(self) -> None:
        shutil.rmtree(self._workspacedir)

    def _build(self, args: tuple[str, ...] | None = None) -> bool:
        return self._run(self.__eclipse_bin,
                            '--launcher.suppressErrors',
                            '-nosplash',
                            '-application', 'org.eclipse.cdt.managedbuilder.core.headlessbuild',
                            '-data', str(self._workspacedir),
                            '-import', str(self._projectdir),
                            '-cleanBuild', f'{self._projectname}/{self._buildtype}',
                            *(args if args is not None else ()),
                        )

    def _copy_buildproduct(self, tag: str) -> None:
        shutil.copy(self._projectdir/self._buildtype/f'{self._projectname}.elf', self._outdir/f'{tag}.elf')

    def _upload(self,
                tag: str,
                logdir: Path,
                args: tuple[str, ...] | None = None,
                cmd: Path | None = None) -> bool:
        cmd = cmd if cmd is not None else self.__upload_bin
        args = args if args is not None else ()
        logger.info('Running: %s %s', cmd, ' '.join(args))
        with (logdir/f'{tag}.txt').open('wb') as logfile:
            _ = logfile.write(' '.join([str(cmd), *args, '\n']).encode('utf-8'))
            returncode, outputs = subprocesstee.run(str(cmd), *args, files={sys.stdout: logfile, sys.stderr: logfile})
        if returncode != 0:
            return False
        if 'Error:' in outputs[1].decode():
            # If output contains the 'Error:' keyword, an error probably happened even though return code may be 0
            return False
        return True

    @abstractmethod
    def prepare(self, tag: str, model: Converter[Any], optimize: str, compression: int) -> Self | None:
        self._create_outdir()
        self._clean_workspace()

        if not self._build():
            return None
        self._copy_buildproduct(tag=tag)
        return self

    @override
    def deploy(self, tag: str) -> Deploy | None:
        logdir = self._outdir/'upload'
        logdir.mkdir(parents=True, exist_ok=True)
        if not self._upload(tag, logdir=logdir):
            return None

        return Deploy(rom_size=self._rom_size(self._outdir/f'{tag}.elf', str(self.__size_bin)),
                      ram_size=self._ram_size(self._outdir/f'{tag}.elf', str(self.__size_bin)),
                      evaluator=self.evaluator)



from __future__ import annotations

import itertools
from datetime import datetime
from pathlib import Path
from typing import Any, NamedTuple

import colorful as cf  # type: ignore[import-untyped]

from qualia_core.qualia import gen_tag, instantiate_model, prepare_deploy
from qualia_core.typing import TYPE_CHECKING
from qualia_core.utils.logger import CSVLogger, Logger

if TYPE_CHECKING:
    from types import ModuleType  # noqa: TCH003

    from qualia_core.datamodel.RawDataModel import RawDataModel  # noqa: TCH001
    from qualia_core.learningframework.LearningFramework import LearningFramework  # noqa: TCH001
    from qualia_core.postprocessing.Converter import Converter  # noqa: TCH001
    from qualia_core.typing import ConfigDict
    from qualia_core.utils import Git  # noqa: TCH001
    from qualia_core.utils.plugin import QualiaComponent  # noqa: TCH001

class PrepareDeployLoggerFields(NamedTuple):
    name: str
    quantize: str
    optimize: str
    compress: int

class PrepareDeploy:
    def __call__(self,  # noqa: PLR0913
                 qualia: QualiaComponent,
                 learningframework: LearningFramework[Any],
                 converter: type[Converter],
                 deployers: ModuleType,
                 data: RawDataModel,
                 config: ConfigDict,
                 git: Git) -> dict[str, Logger[PrepareDeployLoggerFields]]:
        loggers: dict[str, Logger[PrepareDeployLoggerFields]] = {}

        log: CSVLogger[PrepareDeployLoggerFields] = CSVLogger('prepare_deploy',
                                                             file=Path('prepare_deploy')/f'{git.short_hash()}_{datetime.now():%Y-%m-%d_%H-%M-%S}.csv')
        loggers['prepare_deploy'] = log
        # Write column names
        log.fields = PrepareDeployLoggerFields

        for i in range(config['bench']['first_run'], config['bench']['last_run']+1):
            for m, q, o, c in itertools.product(config['model'],
                                                config['deploy']['quantize'],
                                                config['deploy']['optimize'],
                                                config['deploy'].get('compress', [1])):
                if m.get('disabled', False):
                    continue

                # Postprocessings can change model name
                model_name = m['name']
                for postprocessing in config.get('postprocessing', []):
                    ppp = {k: v for k,v in postprocessing.get('params', {}).items()} # Workaround tomlkit bug where some nested dict would lose their items
                    pp = getattr(qualia.postprocessing, postprocessing['kind'])(**ppp)
                    model_name = pp.process_name(model_name)
                    learningframework = pp.process_framework(learningframework)

                # Instantiate model
                model = instantiate_model(dataset=data.sets.test,
                                          framework=learningframework,
                                          model=getattr(learningframework.learningmodels, m['kind']),
                                          model_params=m.get('params', {}),
                                          model_name=model_name,
                                          iteration=i,
                                          load=False, # Model params will be loaded after postprocessings
                                          )

                # Postprocessings can change model topology with PyTorch, needs to be done after instantiating model with new name
                for postprocessing in config.get('postprocessing', []):
                    pp = getattr(qualia.postprocessing, postprocessing['kind'])(**postprocessing.get('params', {}))
                    model, m = pp.process_model(model, m)

                # Show model architecture
                learningframework.summary(model)

                # Load weights after topology optionally changed
                model = learningframework.load(f'{model_name}_r{i}', model)

                print(f'{cf.bold}Preparing {cf.blue}{model_name}{cf.close_fg_color}, run {cf.red}{i}{cf.reset}')
                r = prepare_deploy(
                               datamodel=data,
                               model_kind=m['kind'],
                               model_name=model_name,
                               model=model,
                               framework=learningframework,
                               iteration=i,
                               quantize=q,
                               optimize=o,
                               compress=c,
                               deploy_target=config['deploy']['target'],
                               tag=gen_tag(model_name, q, o, i, c),
                               converter=converter,
                               converter_params=config['deploy'].get('converter', {}).get('params', {}),
                               deployers=deployers,
                               deployer_params=config['deploy'].get('deployer', {}).get('params', {}),
                               representative_dataset=data.sets.train.x)
                if not r:
                    continue

                log(PrepareDeployLoggerFields(name=model_name, quantize=q, optimize=o, compress=c))

        return loggers



from .CIFAR10 import CIFAR10
from .CORe50 import CORe50
from .EllcieHAR import EllcieHAR
from .GSC import GSC
from .GTSRB import GTSRB
from .HD import HD
from .UCI_HAR import UCI_HAR
from .WSMNIST import WSMNIST

__all__ = ['CIFAR10', 'CORe50', 'EllcieHAR', 'GSC', 'GTSRB', 'HD', 'UCI_HAR', 'WSMNIST']



from .Linux import Linux
from .LonganNano import LonganNano
from .NucleoL452REP import NucleoL452REP
from .SparkFunEdge import SparkFunEdge

__all__ = ['Linux', 'LonganNano', 'NucleoL452REP', 'SparkFunEdge']



from __future__ import annotations

import sys
from typing import cast

import torch

from .DataAugmentationPyTorch import DataAugmentationPyTorch

if sys.version_info >= (3, 12):
    from typing import override
else:
    from typing_extensions import override


class TimeShifting(DataAugmentationPyTorch):
    def __init__(self,
                 alpha: float = 1.0,
                 evaluate: bool = False,  # noqa: FBT001, FBT002
                 before: bool = False,  # noqa: FBT001, FBT002
                 after: bool = True) -> None:  # noqa: FBT001, FBT002
        super().__init__(evaluate=evaluate, before=before, after=after)

        self.alpha = alpha

        self.dist = None

    def apply(self, x: torch.Tensor, y: torch.Tensor, device: torch.device) -> tuple[torch.Tensor, torch.Tensor]:
        if self.dist is None:
            self.dist = torch.distributions.beta.Beta(torch.tensor(self.alpha, device=device),
                                                      torch.tensor(self.alpha, device=device))

        ratio = self.dist.sample()

        shift = (x.shape[-1] * ratio).int()

        shifted_x = torch.roll(x, cast(int, shift.item()), dims=-1)

        return shifted_x, y

    @override
    def __call__(self, data: tuple[torch.Tensor, torch.Tensor], device: torch.device) -> tuple[torch.Tensor, torch.Tensor]:
        return self.apply(*data, device=device)



from __future__ import annotations

import copy
import logging
from collections.abc import Mapping
from typing import Any, TypeVar, Union, cast

logger = logging.getLogger(__name__)

RecursiveConfigUnion = Union[Any, Mapping[str, Any], list[Any]]
RecursiveMapping = Mapping[str, RecursiveConfigUnion]

T = TypeVar('T', str, int)
TRecursiveDict = dict[T, RecursiveConfigUnion]
TRecursiveMapping = Mapping[T, RecursiveConfigUnion]

U = TypeVar('U', bound=RecursiveMapping)

def merge_dict(dict_a: U,
               dict_b: U,
               merge_lists: bool = False) -> U:  # noqa: FBT001, FBT002
    """Merge elements from dict_b into dict_a if they are missing."""
    def merger(table: TRecursiveDict[T],
               template: TRecursiveMapping[T]) -> None:
        for k, template_v in template.items():
            if k not in table:
                table[k] = copy.deepcopy(template_v)
                continue

            table_v = table[k]
            if isinstance(template_v, dict):
                if not isinstance(table_v, dict):
                    logger.error("Incompatible target value type %s for key '%s', expected dict", type(table_v), k)
                    raise TypeError
                merger(table_v, template_v)
            elif isinstance(template_v, list) and merge_lists:
                if not isinstance(table_v, dict):
                    logger.error("Incompatible target value type %s for key '%s', expected dict", type(table_v), k)
                    raise TypeError
                # Convert possible string keys to int
                temp_table = {int(tblk): tblv for tblk, tblv in table_v.items()}
                # Convert template list to dict
                v_dict: TRecursiveDict[int] = dict(enumerate(template_v))
                # Merge converted template dict into table
                merger(temp_table, v_dict)
                # Convert back dest table to list, iterate up to highest index
                table[k] = [temp_table[tbli] for tbli in range(sorted(temp_table.keys())[-1] + 1)]

    # Mapping becomes Dict after copy as it is mutable and modified by merger()
    dict_r = cast(TRecursiveDict[str], copy.deepcopy(dict_a))
    merger(dict_r, dict_b)
    # Merged dict is cast back to original Mapping type to keep type compatibility
    return cast(U, dict_r)




import git

class Git:
    def __init__(self):
        self.repo = git.Repo(search_parent_directories=True)

    def short_hash(self):
        sha = self.repo.head.commit.hexsha
        return self.repo.git.rev_parse(sha, short=8)



from __future__ import annotations

import logging
import os
import time
from dataclasses import astuple, dataclass
from pathlib import Path
from typing import Any

import blosc2
import numpy as np
import numpy.typing

from .DataModel import DataModel

logger = logging.getLogger(__name__)

@dataclass
class RawData:
    x: numpy.typing.NDArray[Any]
    y: numpy.typing.NDArray[Any]
    info: numpy.typing.NDArray[Any] | None = None

    @property
    def data(self) -> numpy.typing.NDArray[Any]:
        return self.x

    @data.setter
    def data(self, data: numpy.typing.NDArray[Any]) -> None:
        self.x = data

    @property
    def labels(self) -> numpy.typing.NDArray[Any]:
        return self.y

    @labels.setter
    def labels(self, labels: numpy.typing.NDArray[Any]) -> None:
        self.y = labels

    def export(self, path: Path, compressed: bool = True) -> None:
        start = time.time()
        if compressed:
            cparams = {'codec': blosc2.Codec.ZSTD, 'clevel': 5, 'nthreads': os.cpu_count()}
            blosc2.pack_array2(np.ascontiguousarray(self.data), urlpath=str(path/'data.npz'), mode='w', cparams=cparams)
            blosc2.pack_array2(np.ascontiguousarray(self.labels), urlpath=str(path/'labels.npz'), mode='w', cparams=cparams)
            if self.info is not None:
                blosc2.pack_array2(np.ascontiguousarray(self.info), urlpath=str(path/'info.npz'), mode='w', cparams=cparams)
        else:
            np.savez(path/'data.npz', data=self.data)
            np.savez(path/'labels.npz', labels=self.labels)
            if self.info is not None:
                np.savez(path/'info.npz', info=self.info)
        logger.info('export() Elapsed: %s s', time.time() - start)

    @classmethod
    def import_data(cls, path: Path, compressed: bool = True) -> RawData | None:
        start = time.time()

        for fname in ['data.npz', 'labels.npz']:
            if not (path/fname).is_file():
                logger.error("'%s' not found. Did you run 'preprocess_data'?", path/fname)
                return None

        info: numpy.typing.NDArray[Any] | None = None

        if compressed:
            data: numpy.typing.NDArray[Any] = blosc2.load_array(str(path/'data.npz'))
            labels: numpy.typing.NDArray[Any] = blosc2.load_array(str(path/'labels.npz'))
            if (path/'info.npz').is_file():
                info = blosc2.load_array(str(path/'info.npz'))
        else:
            with np.load(path/'data.npz') as datanpz:
                data = datanpz['data']
            with np.load(path/'labels.npz') as labelsnpz:
                labels = labelsnpz['labels']

            if (path/'info.npz').is_file():
                with np.load(path/'info.npz') as infonpz:
                    info = infonpz['info']

        ret = cls(x=data, y=labels, info=info)
        logger.info('import_data() Elapsed: %s s', time.time() - start)
        return ret

    def astuple(self) -> tuple[Any, ...]:
        return astuple(self)


class RawDataSets(DataModel.Sets[RawData]):
    ...


class RawDataModel(DataModel[RawData]):
    sets: DataModel.Sets[RawData]

    @classmethod
    def import_data(cls, name: str, sets: list[str] | None = None) -> RawDataModel | None:
        set_names = sets if sets is not None else list(RawDataSets.fieldnames())
        sets_dict: dict[str, RawData | None] =  {sname: RawData.import_data(Path('out')/'data'/name/sname)
                                                   for sname in set_names}

        if any(s is None for s in sets_dict.values()):
            logger.error('Could not import data.')
            return None

        logger.info('Imported %s for %s', ', '.join(sets_dict.keys()), name)

        return cls(sets=RawDataSets(**sets_dict),
                   name=name)



from __future__ import annotations

from dataclasses import dataclass

from qualia_core.typing import TYPE_CHECKING

# We are inside a TYPE_CHECKING block but our custom TYPE_CHECKING constant triggers TCH001-TCH003 so ignore them
if TYPE_CHECKING:
    from qualia_core.evaluation.Evaluator import Evaluator  # noqa: TCH001


@dataclass
class Deploy:
    rom_size: int | None
    ram_size: int | None
    evaluator: type[Evaluator]




from __future__ import annotations

import logging
import sys
from pathlib import Path
from typing import TYPE_CHECKING

from qualia_core.experimenttracking.ExperimentTracking import ExperimentTracking
from qualia_core.utils.file.DirectoryReader import DirectoryReader

if sys.version_info >= (3, 12):
    from typing import override
else:
    from typing_extensions import override

if TYPE_CHECKING:
    from qualia_core.typing import RecursiveConfigDict

logger = logging.getLogger(__name__)

class ClearML(ExperimentTracking):
    def __init__(self,
                 project_name: str,
                 task_name: str,
                 sources_path: str | None = None,
                 ignores: list[str] | None = None,
                 offline_mode: bool = False) -> None:
        from clearml import Task

        super().__init__()
        self.__project_name = project_name
        self.__task_name = task_name
        self.__task: Task | None = None

        # Path to the current project sources, used as the "script path" and to find the version control repository (e.g. git)
        # By default use the "src/qualia" path under the current working directory
        src_path = Path(sources_path) if sources_path is not None else Path.cwd() / 'src' / 'qualia'

        # Paths to ignore when looking for requirements
        # Parsing everything would take much longer than necessary
        # This mostly leaves the src/ and tests/ directories (and conf)
        ignores_list = ['__pypackages__',
                        'logs',
                        'out',
                        'lightning_logs',
                        'build',
                        'third_party',
                        '.venv',
                        '.mypy_cache',
                        '.ruff_cache',
                        '.pyre',
                        '.pytest_cache',
                        'data',
                        'src/qualia/assets']
        if ignores is not None:
            ignores_list += ignores

        if offline_mode:
            Task.set_offline(offline_mode=offline_mode)

        self.__patch_clearml(extra_ignores=ignores_list, src_path=src_path)

    def __patch_clearml(self, extra_ignores: list[str], src_path: Path) -> None:
        """Monkey-patch ClearML methods for our use-case.

        ScriptInfo.get() patched to set the desired 'script_path' to find the git repository.
        pigar.GenerateReqs.__init__() patched to add more directories to the ignore list of get_requirements().
        """
        from clearml.backend_interface.task.repo import scriptinfo
        getScriptInfo = scriptinfo.ScriptInfo.get.__func__

        def getScriptInfoPatched(cls, filepaths=None, *args, **kwargs):
            return getScriptInfo(cls, *args, filepaths=[src_path], **kwargs)

        scriptinfo.ScriptInfo.get = classmethod(getScriptInfoPatched)

        import clearml.utilities.pigar.__main__ as pigar
        generateReqsInit = pigar.GenerateReqs.__init__
        def generateReqsInitPatched(self: pigar.GenerateReqs, *, ignores: list[str], **kwargs) -> None:
            generateReqsInit(self, ignores=ignores + extra_ignores, **kwargs)
        pigar.GenerateReqs.__init__ = generateReqsInitPatched


    @override
    def start(self, name: str | None = None) -> None:
        from clearml import Task
        if Task.current_task() is None:
            task_name = f'{self.__task_name}_{name}' if name is not None else self.__task_name
            self.__task = Task.init(project_name=self.__project_name, task_name=task_name, reuse_last_task_id=False)

    @override
    def stop(self) -> None:
        if self.__task is not None:
            self.__task.close()
        self.__task = None

    @override
    def _hyperparameters(self, params: RecursiveConfigDict) -> None:
        if self.__task is not None:
            self.__task.connect(params)

    @override
    @classmethod
    def initializer(cls) -> None:
        """Connect to task in PyTorch Lightning Trainer subprocess with e.g. ddp_spawn."""
        from clearml import Task
        Task.current_task()

    @classmethod
    def import_and_clear_all_offline_sessions(cls) -> None:
        from clearml import Task
        from qualia_core.main import setup_logging

        # We main not be called from qualia_core.main:main so always setup logging to show logger.info()
        setup_logging()

        offline_cache_path = Path(sys.argv[1]) if len(sys.argv) > 0 else Path.home()/'.clearml'/'cache'/'offline'

        dr = DirectoryReader()
        sessions = list(dr.read(directory=offline_cache_path, ext='.zip', recursive=False))
        imported_sessions: list[str] = []
        logger.info('%d session%s to import.', len(sessions), 's' if len(sessions) > 1 else '')

        for f in sessions:
            imported_session = Task.import_offline_session(session_folder_zip=str(f))
            if imported_session is None:
                logger.error('Failed to import session %s', str(f))
            else:
                imported_sessions.append(imported_session)

            logger.info('Deleting session archive %s', str(f))
            f.unlink()

        logger.info('Imported %d session%s.', len(imported_sessions), 's' if len(imported_sessions) > 1 else '')

    @property
    def logger(self) -> None:
        return None



from __future__ import annotations

import inspect
import sys
from dataclasses import dataclass
from typing import Any, TypeVar

import colorful as cf  # type: ignore[import-untyped]

from qualia_core.learningmodel.LearningModel import LearningModel
from qualia_core.typing import TYPE_CHECKING, ModelParamsConfigDict, OptimizerConfigDict

if TYPE_CHECKING:
    from qualia_core.dataaugmentation.DataAugmentation import DataAugmentation  # noqa: TCH001
    from qualia_core.datamodel.DataModel import DataModel  # noqa: TCH001
    from qualia_core.datamodel.RawDataModel import RawData, RawDataModel  # noqa: TCH001
    from qualia_core.experimenttracking.ExperimentTracking import ExperimentTracking  # noqa: TCH001
    from qualia_core.learningframework.LearningFramework import LearningFramework  # noqa: TCH001
    from qualia_core.typing import RecursiveConfigDict

@dataclass
class TrainResult:
    name: str
    i: int
    model: Any
    params: int
    mem_params: int
    acc: float
    metrics: dict[str, Any]
    datamodel: DataModel[RawData]
    trainset: RawData
    testset: RawData
    framework: LearningFramework[Any]
    batch_size: int
    optimizer: Any
    log: bool
    dataaugmentations: list[DataAugmentation]
    experimenttracking: ExperimentTracking | None

T = TypeVar('T', bound=LearningModel)

def gen_tag(mname: str, q: str, o: int, i: int, c: int) -> str:
    return f'{mname}_q{q}_o{o}_c{c}_r{i}'

def instantiate_model(dataset: RawData,  # noqa: PLR0913
                      framework: LearningFramework[T],
                      model: type[T],
                      model_params: ModelParamsConfigDict | None,
                      model_name: str,
                      iteration: int,
                      load: bool = True) -> T:  # noqa: FBT001, FBT002
    model_params = model_params if model_params is not None else ModelParamsConfigDict()

    if 'input_shape' not in model_params:
        model_params['input_shape'] = dataset.x.shape[1:]
    else:
        model_params['input_shape'] = tuple(model_params['input_shape'])
    if 'output_shape' not in model_params:
        model_params['output_shape'] = dataset.y.shape[1:]
    else:
        model_params['output_shape'] = tuple(model_params['output_shape'])

    if 'iteration' in inspect.signature(model).parameters:
        model_params['iteration'] = iteration

    # Instantiate model
    new_model = model(**model_params)

    if load:
        new_model = framework.load(f'{model_name}_r{iteration}', new_model)

    print(f'{new_model.input_shape=} {new_model.output_shape=}')

    # Show model architecture
    framework.summary(new_model)

    return new_model

def train(datamodel: RawDataModel,  # noqa: PLR0913
          train_epochs: int,
          iteration: int,
          framework: LearningFramework[T],
          model: type[T],
          model_name: str,
          model_params: RecursiveConfigDict | None = None,
          batch_size: int | None = None,
          optimizer: OptimizerConfigDict | None = None,
          load: bool = False,  # noqa: FBT002, FBT001
          train: bool = True,  # noqa: FBT001, FBT002
          evaluate: bool = True,  # noqa: FBT001, FBT002
          dataaugmentations: list[DataAugmentation] | None = None,
          experimenttracking: ExperimentTracking | None = None) -> TrainResult:

    if batch_size is None:
        batch_size = 32

    new_model = instantiate_model(dataset=datamodel.sets.train,
                       framework=framework,
                       model=model,
                       model_params=model_params,
                       model_name=model_name,
                       iteration=iteration,
                       load=load)

    # Export model visualization to dot file
    framework.save_graph_plot(new_model, f'{model_name}_r{iteration}')

    # You can plot the quantize training graph on tensorboard
    if train:
        new_model = framework.train(new_model,
                        trainset=datamodel.sets.train,
                        validationset=datamodel.sets.valid,
                        epochs=train_epochs,
                        batch_size=batch_size,
                        optimizer=optimizer,
                        dataaugmentations=dataaugmentations,
                        experimenttracking=experimenttracking,
                        name=f'{model_name}_r{iteration}_train')

    metrics = {}
    if evaluate:
        print(f'{cf.bold}Evaluation on train dataset{cf.reset}')
        metrics = framework.evaluate(new_model,
                                 datamodel.sets.train,
                                 batch_size=batch_size,
                                 dataaugmentations=dataaugmentations,
                                 experimenttracking=experimenttracking,
                                 dataset_type='train',
                                 name=f'{model_name}_r{iteration}_eval_train')

        if len(datamodel.sets.test.x) > 0: # Don't evaluate if testset is empty
            print(f'{cf.bold}Evaluation on test dataset{cf.reset}')
            metrics = framework.evaluate(new_model,
                                     datamodel.sets.test,
                                     batch_size=batch_size,
                                     dataaugmentations=dataaugmentations,
                                     experimenttracking=experimenttracking,
                                     dataset_type='test',
                                     name=f'{model_name}_r{iteration}_eval_test')

    # Do not save loaded model that hasn't been retrained
    if train or not load:
        framework.export(new_model, f'{model_name}_r{iteration}')

    return TrainResult(name=model_name,
                       i=iteration,
                       model=new_model,
                       params=framework.n_params(new_model),
                       mem_params=framework.n_params(new_model) * 4, # Non-quantized model is assumed to be 32 bits
                       acc=metrics.get('testacc', None),
                       metrics=metrics,
                       datamodel=datamodel,
                       trainset=datamodel.sets.train,
                       testset=datamodel.sets.test,
                       framework=framework,
                       batch_size=batch_size,
                       optimizer=optimizer,
                       log=True,
                       dataaugmentations=dataaugmentations,
                       experimenttracking=experimenttracking)

def prepare_deploy(
    datamodel,
    model_kind,
    model_name,
    model,
    framework,
    iteration,
    deploy_target,
    quantize='float32',
    optimize=None,
    compress=1,
    tag='main',
    converter=None,
    converter_params={},
    deployers=None,
    deployer_params={},
    representative_dataset=None):

    if not converter: # no custom converter passed as parameter, check if model suggests custom converter
        converter = getattr(model_kind, 'converter', False)

    if converter:
        ca = converter(quantize=quantize, **converter_params).convert(framework, model, f'{model_name}_r{iteration}', representative_dataset=representative_dataset)
    else: # No conversion taking place since no converter was specified
        ca = model

    if ca is None:
        return None

    if not deployers:
        if not converter:
            print('Error: no converter and no deployers specified', file=sys.stderr)
            return None
        elif not hasattr(ca, 'deployers'):
            print('Error: no deployers specified and converter does not suggest a deployer', file=sys.stderr)
            return None
        else:
            deployers = ca.deployers
    return getattr(deployers, deploy_target)(**deployer_params).prepare(tag=tag, model=ca, optimize=optimize, compression=compress)

def deploy(model_kind, deploy_target, tag='main', deployers=None, deployer_params={}):
    if not deployers: # no custom deployers passed as parameter, check if model suggests custom converter
        converter = getattr(model_kind, 'converter', False)
        if converter and converter.deployers: # Converter suggested deployers
            deployers = converter.deployers

    return getattr(deployers, deploy_target)(**deployer_params).deploy(tag=tag)

def evaluate(
    datamodel,
    model_kind,
    model_name,
    model,
    framework,
    iteration,
    target,
    quantization,
    fmem_params,
    tag,
    limit=None,
    evaluator=None,
    evaluator_params={},
    dataaugmentations=None):

    if not evaluator: # no custom deployers passed as parameter, check if model suggests custom converter
        converter = getattr(model_kind, 'converter', False)
        if converter and converter.evaluator: # Converter suggested deployers
            evaluator = converter.evaluator

    if not evaluator:
        raise ValueError('No evaluator')
    result = evaluator(**evaluator_params).evaluate(framework=framework,
                                                    model_kind=model_kind,
                                                    dataset=datamodel,
                                                    target=target,
                                                    tag=tag,
                                                    limit=limit,
                                                    dataaugmentations=[da for da in dataaugmentations if da.evaluate])
    if not result:
        return result

    # fill in iteration, name quantization from context
    result.name = model_name
    result.i = iteration
    result.quantization = quantization
    # fill in params count and memory from model
    result.params = framework.n_params(model)
    result.mem_params = fmem_params(framework, model)

    return result
